# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os

from . import pixel_scale, n_pix, pixel_scale_mm
from . import n_pix, n_sca, longwave_bands, shortwave_bands
from . import diameter, obscuration
from .roman_bandpass import getBandpasses

from ..utilities import LRU_Cache
from ..position import PositionD
from ..errors import GalSimValueError, GalSimRangeError
from ..bandpass import Bandpass
from ..wcs import PixelScale
from ..gsparams import GSParams
from ..phase_psf import Aperture
from .. import OpticalPSF, ChromaticOpticalPSF
from .. import fits
from .. import meta_data


"""
@file roman_psfs.py

Part of the Roman Space Telescope module.  This file includes routines needed to define a realistic
PSF for Roman.
"""

# Define a default set of bandpasses for which this routine works.
default_bandpass_list = ['J129', 'F184', 'W149', 'Y106', 'Z087', 'H158']
# Prefix for files containing information about Zernikes for each SCA for cycle 7.
zemax_filepref = "Roman_Cycle-9_WFI_zim_zernikes_082623"
zemax_filesuff = '.txt'
zemax_wavelength = 1293. #nm

# These need 'SCA*' prepended to the start to get the file name, and they live in
# the share/roman directory.
pupil_plane_file_longwave = 'RST_WIM_Filter_F184_SCA_'
pupil_plane_file_shortwave = 'RST_WIM_Filter_skinny_SCA_'
pupil_plane_filesuff = '.fits.gz'
def getPSF(SCA, bandpass,
           SCA_pos=None, pupil_bin=4, wcs=None,
           n_waves=None, extra_aberrations=None,
           wavelength=None, gsparams=None,
           logger=None, high_accuracy=None, approximate_struts=None):
    """Get a single PSF for Roman ST observations.

    The user must provide the SCA and bandpass; the latter is used when setting up the pupil
    plane configuration and when interpolating chromatic information, if requested.

    This routine carries out linear interpolation of the aberrations within a given SCA, based on
    the Roman Cycle 9 specification of the aberrations as a function of focal plane
    position, more specifically from the WebbPSF data files from webbpsf-data-1.2.1.tar.gz
    downloaded from https://webbpsf.readthedocs.io/en/latest/installation.html#data-install.
    The abberation file is webbpsf-data/WFI/wim_zernikes_cycle9.csv.

    The mask images for the Roman pupil plane are available in the same WebbPSF data files.
    There are separate files for each SCA, since the view of the spider pattern varies somewhat
    across the field of view of the wide field camera. Users usually don't need
    to worry about any of this, as GalSim will select the correct pupil image automatically based
    on the SCA and bandpass provided.

    The full pupil plane images are 4096 x 4096, which use a lot of memory and are somewhat slow
    to use, so we normally bin them by a factor of 4 (resulting in 1024 x 1024 images). This
    provides enough detail for most purposes and is much faster to render than using the full pupil
    plane images.  This bin factor is a settable parameter, called ``pupil_bin``.  If you want the
    more accurate, slower calculation using the full images, you can set it to 1. In the other
    direction, using pupil_bin=8 (resulting in a 512 x 512 image) still provides fairly reasonable
    results and is even faster to render.  It is not generally recommended to use higher binning
    than that, as the diffraction spikes will become noticeably degraded.

    .. note::

        This function will cache the aperture calculation, so repeated calls with the same
        SCA and bandpass should be much faster after the first call, as the pupil plane will
        already be loaded.  If you need to clear the cache for memory reasons, you may call::

            galsim.roman.roman_psfs._make_aperture.clear()

        to recover any memory currently being used for this cache.  Of course, subsequent calls to
        `getPSF` will need to rebuild the aperture at that point.

    The PSF that is returned by default will be oriented with respect to the SCA coordinates,
    not world coordinates as is typical in GalSim.  The pupil plane has a fixed orientation
    with respect to the focal plane, so the PSF rotates with the telescope.  To obtain a
    PSF in world coordinates, which can be convolved with galaxies (that are normally described
    in world coordinates), you may pass in a ``wcs`` parameter to this function.  This will
    project the PSF into world coordinates according to that WCS before returning it.  Otherwise,
    the return value is equivalent to using ``wcs=galim.PixelScale(galsim.roman.pixel_scale)``.

    The calculation takes advantage of the fact that the diffraction limit and aberrations have a
    simple, understood wavelength-dependence.  (The Roman abberation data for Cycle 9 does in fact
    provide aberrations as a function of wavelength, but the deviation from the expected chromatic
    dependence is sub-percent so we neglect it here.)  For reference, the script used to parse the
    Zernikes given on the webpage and create the files in the GalSim repository can be found in
    ``devel/external/parse_roman_zernikes_1217.py``.  The resulting chromatic object can be used to
    draw into any of the Roman bandpasses, though the pupil plane configuration will only be
    correct for those bands in the same range (i.e., long- or short-wavelength bands).

    For applications that require very high accuracy in the modeling of the PSF, with very limited
    aliasing, you may want to lower the folding_threshold in the gsparams.  Otherwise very bright
    stars will show some reflections in the spider pattern and possibly some boxiness at the
    outskirts of the PSF.  Using ``gsparams = GSParams(folding_threshold=2.e-3)`` generally
    provides good results even for very bright (e.g. mag=10) stars.  In these cases, you probably
    also want to reduce ``pupil_bin`` somewhat from the default value of 4.

    By default, no additional aberrations are included above the basic design.  However, users can
    provide an optional keyword ``extra_aberrations`` that will be included on top of those that are
    part of the design.  This should be in the same format as for the ChromaticOpticalPSF class,
    with units of waves at the fiducial wavelength, 1293 nm. Currently, only aberrations up to order
    22 (Noll convention) are simulated.  For Roman, the tolerance for additional
    aberrations was a total of 90 nanometers RMS as of mid-2015, distributed largely among coma,
    astigmatism, trefoil, and spherical aberrations (NOT defocus).  This information might serve as
    a guide for reasonable ``extra_aberrations`` inputs.  The reference for that number is
    an earlier Cycle 5 document:

    http://roman.gsfc.nasa.gov/science/sdt_public/wps/references/instrument/README_AFTA_C5_WFC_Zernike_and_Field_Data.pdf

    However, the default (non-extra) aberrations are from Cycle 7 material linked earlier in this
    docstring.

    Jitter and charge diffusion are, by default, not included.  Users who wish to include these can
    find some guidelines for typical length scales of the Gaussians that can represent these
    effects, and convolve the ChromaticOpticalPSF with appropriate achromatic Gaussians.

    The PSFs are always defined assuming the user will specify length scales in arcsec.

    Users may find they do not have to call `getPSF` for all objects in their simulations; for a
    given SCA and position within the SCA, and a given pupil plane configuration and wavelength
    information, it should be possible to reuse the PSFs.

    Parameters:
        SCA:                Single value specifying the SCA for which the PSF should be
                            loaded.
        bandpass:           Single string specifying the bandpass to use when defining the
                            pupil plane configuration and/or interpolation of chromatic PSFs.
                            You may also pass a string 'long' or 'short' for this argument, in
                            which case, the correct pupil plane configuration will be used for
                            long- or short-wavelength bands (F184 is long, all else is short).
                            In this case, no interpolation can be used, since it is defined
                            using the extent of the chosen bandpass. If ``wavelength`` is given,
                            then bandpass may be None, which will use the short-wavelength pupil
                            plane image.
        SCA_pos:            Single galsim.PositionD indicating the position within the SCA
                            for which the PSF should be created. If None, the exact center of
                            the SCA is chosen. [default: None]
        pupil_bin:          The binning to apply to the pupil plane image. (See discussion above.)
                            [default: 4]
        wcs:                The WCS to use to project the PSF into world coordinates.
                            [default: galsim.PixelScale(galsim.roman.pixel_scale)]
        n_waves:            Number of wavelengths to use for setting up interpolation of the
                            chromatic PSF objects, which can lead to much faster image
                            rendering.  If None, then no interpolation is used. Note that
                            users who want to interpolate can always set up the interpolation
                            later on even if they do not do so when calling `getPSF`.
                            [default: None]
        extra_aberrations:  Array of extra aberrations to include in the PSF model, on top of
                            those that are part of the Roman design.  These should be
                            provided in units of waves at the fiducial wavelength of 1293 nm,
                            as an array of length 23 with entries 4 through 22 corresponding
                            to defocus through the 22nd Zernike in the Noll convention.
                            [default: None]
        wavelength:         An option to get an achromatic PSF for a single wavelength, for
                            users who do not care about chromaticity of the PSF.  If None,
                            then the fully chromatic PSF is returned.  Alternatively the user
                            should supply either (a) a wavelength in nanometers, and they
                            will get achromatic OpticalPSF objects for that wavelength, or
                            (b) a bandpass object, in which case they will get achromatic
                            OpticalPSF objects defined at the effective wavelength of that
                            bandpass.  [default: False]
        gsparams:           An optional GSParams argument.  See the docstring for GSParams
                            for details. [default: None]

    Returns:
        A single PSF object (either a ChromaticOpticalPSF or an OpticalPSF depending on the
        inputs).

    """

    # Deprecated options
    if bandpass == 'W149':
        from ..deprecated import depr
        depr('W149', 2.5, 'W146', 'Note: this is to match current Roman filter naming schemes')
        bandpass = 'W146'
    if high_accuracy:
        if approximate_struts:
            from ..deprecated import depr
            depr('high_accuracy=True,approximate_struts=True', 2.3,
                 'pupil_bin=4, gsparams=galsim.GSParams(folding_threshold=2.e-3)',
                 'Note: this is not actually equivalent to the old behavior, but it should '
                 'be both faster and more accurate than the corresponding PSF in v2.2.')
            # Set folding_threshold 2.5x smaller than default.
            gsparams = GSParams.check(gsparams, folding_threshold=2.e-3)
            pupil_bin = 4
        else:
            from ..deprecated import depr
            depr('high_accuracy=True', 2.3,
                 'pupil_bin=1, gsparams=galsim.GSParams(folding_threshold=2.e-3)',
                 'Note: this is not actually equivalent to the old behavior, but it should '
                 'be both faster and more accurate than the corresponding PSF in v2.2.')
            # Set folding_threshold 2.5x smaller than default.
            gsparams = GSParams.check(gsparams, folding_threshold=2.e-3)
            pupil_bin = 1
    elif approximate_struts:
        from ..deprecated import depr
        depr('approximate_struts=True', 2.3, 'pupil_bin=8',
             'Note: this is not actually equivalent to the old behavior, but it should '
             'be both faster and more accurate than the corresponding PSF in v2.2.')
        pupil_bin = 8
    elif approximate_struts is False or high_accuracy is False:
        # If they are explicitly given, rather than default (None), then trigger this.
        from ..deprecated import depr
        depr('approximate_struts=False, high_accuracy=False', 2.3, 'pupil_bin=4',
             'Note: this is not actually equivalent to the old behavior, but it should '
             'be both faster and more accurate than the corresponding PSF in v2.2.')
        pupil_bin = 4

    if SCA <= 0 or SCA > n_sca:
        raise GalSimRangeError("Invalid SCA.", SCA, 1, n_sca)

    # SCA_pos: if None, then all should just be center of the SCA.
    if SCA_pos is None:
        SCA_pos = PositionD(n_pix/2, n_pix/2)

    # Parse the bandpasses to see which pupil plane image is needed
    pupil_plane_type = None
    if bandpass in longwave_bands or bandpass=='long':
        pupil_plane_type = 'long'
    elif bandpass in shortwave_bands or bandpass=='short':
        pupil_plane_type = 'short'
    elif bandpass is None and n_waves is None:
        pupil_plane_type = 'short'
    else:
        raise GalSimValueError("Bandpass not a valid Roman bandpass or 'short'/'long'.",
                               bandpass, default_bandpass_list)

    # If bandpass is 'short'/'long', then make sure that interpolation is not called for, since that
    # requires an actual bandpass.
    if bandpass in ['short','long'] and n_waves is not None:
        raise GalSimValueError("Cannot use bandpass='short'/'long' with interpolation.", bandpass)

    if not isinstance(wavelength, (Bandpass, float, type(None))):
        raise TypeError("wavelength should either be a Bandpass, float, or None.")

    # Now call _get_single_PSF().
    psf = _get_single_PSF(SCA, bandpass, SCA_pos, pupil_bin,
                          n_waves, extra_aberrations, wavelength,
                          pupil_plane_type, gsparams)

    # Apply WCS.
    # The current version is in arcsec units, but oriented parallel to the image coordinates.
    # So to apply the right WCS, project to pixels using the Roman mean pixel_scale, then
    # project back to world coordinates with the provided wcs.
    if wcs is not None:
        scale = PixelScale(pixel_scale)
        psf = wcs.toWorld(scale.toImage(psf), image_pos=SCA_pos)

    return psf

def __make_aperture(SCA, pupil_plane_type, pupil_bin, wave, gsparams):
    # Load the pupil plane image.
    if pupil_plane_type == 'long':
        pupil_plane_im = os.path.join(meta_data.share_dir, 'roman',
            pupil_plane_file_longwave + '%d'%SCA + pupil_plane_filesuff)
    else:
        pupil_plane_im = os.path.join(meta_data.share_dir, 'roman',
            pupil_plane_file_shortwave + '%d'%SCA + pupil_plane_filesuff)
    pupil_plane_im = fits.read(pupil_plane_im, read_header=True)
    # Native pixel scale in the file is for the exit pupil.  We want the scale of the
    # entrance pupil.  Fortunately, they provide the conversion as 'HIERARCH PUPIL SCALE FACTOR' in the header.
    # They also use microns for units, and we want meters, hence the extra 1.e-6.
    pupil_plane_im.scale *= pupil_plane_im.header['HIERARCH PUPIL SCALE FACTOR'] * 1.e-6

    pupil_plane_im = pupil_plane_im.bin(pupil_bin,pupil_bin)

    aper = Aperture(lam=wave, diam=diameter,
                    obscuration=obscuration,
                    pupil_plane_im=pupil_plane_im,
                    gsparams=gsparams)
    return aper

# Usually a given run will only need one or a few different apertures for repeated getPSF calls.
# So cache those apertures here to avoid having to remake them.
_make_aperture = LRU_Cache(__make_aperture)

def _get_single_PSF(SCA, bandpass, SCA_pos, pupil_bin,
                    n_waves, extra_aberrations, wavelength,
                    pupil_plane_type, gsparams):
    """Routine for making a single PSF.  This gets called by `getPSF` after it parses all the
       options that were passed in.  Users will not directly interact with this routine.
    """
    if wavelength is None:
        wave = zemax_wavelength
    elif isinstance(wavelength, Bandpass):
        wave = wavelength = wavelength.effective_wavelength
    else:
        wave = wavelength

    # All parameters relevant to the aperture.  We may be able to use a cached version.
    aper = _make_aperture(SCA, pupil_plane_type, pupil_bin, wave, gsparams)

    # Start reading in the aberrations for that SCA
    aberrations, x_pos, y_pos = _read_aberrations(SCA)
    # Do bilinear interpolation, unless we're exactly at the center (default).
    use_aberrations = _interp_aberrations_bilinear(aberrations, x_pos, y_pos, SCA_pos)

    if extra_aberrations is not None:
        use_aberrations[:len(extra_aberrations)] += extra_aberrations
    # We don't want to use piston, tip, or tilt aberrations.  The former doesn't affect the
    # appearance of the PSF, and the latter cause centroid shifts.  So, we set the first 4
    # numbers (corresponding to a place-holder, piston, tip, and tilt) to zero.
    use_aberrations[0:4] = 0.

    # Now set up the PSF, including the option to interpolate over waves
    if wavelength is None:
        PSF = ChromaticOpticalPSF(lam=zemax_wavelength,
                                  diam=diameter, aberrations=use_aberrations,
                                  aper=aper, gsparams=gsparams)
        if n_waves is not None:
            # To decide the range of wavelengths to use, check the bandpass.
            bp_dict = getBandpasses()
            bp = bp_dict[bandpass]
            PSF = PSF.interpolate(waves=np.linspace(bp.blue_limit, bp.red_limit, n_waves),
                                  oversample_fac=1.5)
    else:
        tmp_aberrations = use_aberrations * zemax_wavelength / wavelength
        PSF = OpticalPSF(lam=wavelength, diam=diameter,
                         aberrations=tmp_aberrations,
                         aper=aper, gsparams=gsparams)

    return PSF

def _read_aberrations(SCA):
    """
    This is a helper routine that reads in aberrations for a particular SCA and wavelength (given as
    galsim.roman.roman_psfs.zemax_wavelength) from stored files, and returns them along with the
    field positions.

    Parameters:
        SCA:        The identifier for the SCA, from 1-18.

    Returns:
        NumPy arrays containing the aberrations, and x and y field positions.
    """
    # Construct filename.
    sca_str = '_%02d'%SCA
    infile = os.path.join(meta_data.share_dir, 'roman',
                          zemax_filepref + sca_str + zemax_filesuff)

    # Read in data.
    dat = np.loadtxt(infile)
    # It actually has 5 field positions, not just 1, to allow us to make position-dependent PSFs
    # within an SCA eventually.  Put it in the required format: an array of length (5 field
    # positions, 23 Zernikes), with the first entry empty (Zernike polynomials are 1-indexed so we
    # use entries 1-22).  The units are waves.
    aberrations = np.zeros((5,23))
    aberrations[:,1:] = dat[:,5:]
    # Also get the field position.  The file gives it in mm with respect to the center, but we
    # want it in pixels with respect to the corner. The pixel size of the detector is 0.01 mm/pixel

    x = dat[:,1]/pixel_scale_mm
    y = dat[:,2]/pixel_scale_mm
    if SCA % 3 != 0:
        # For these, the SCA is rotated 180 degrees relative to the nominal X, Y coordinates.
        x = -x
        y = -y

    x += n_pix/2
    y += n_pix/2

    return aberrations, x, y

def _interp_aberrations_bilinear(aberrations, x_pos, y_pos, SCA_pos):
    """
    This is a helper routine to do bilinear interpolation of aberrations defined at 4 field
    positions: the four corners.  Note that we also have aberrations at the center position,
    but these are generally quite close (within a few percent) of what would come from this bilinear
    interpolation.  So for simplicity, we just do the bilinear interpolation.
    """
    # The data comprise 5 rows: center, lower-left, upper-left, upper-right, lower-right.
    # The x,y values at the corners aren't precisely identical, but despite that, just
    # take the outer one and do a simple bilinear interpolation as though it were a rectangle.

    # First, figure out which point is which corner. (0 is always the center.)
    for i in range(1,5):
        if x_pos[i] < x_pos[0] and y_pos[i] < y_pos[0]:
            ll = i # lower-left
        if x_pos[i] < x_pos[0] and y_pos[i] > y_pos[0]:
            ul = i # upper-left
        if x_pos[i] > x_pos[0] and y_pos[i] < y_pos[0]:
            lr = i # lower-right
        if x_pos[i] > x_pos[0] and y_pos[i] > y_pos[0]:
            ur = i # upper-right
    assert x_pos[ll] < x_pos[0] and y_pos[ll] < y_pos[0]
    assert x_pos[ul] < x_pos[0] and y_pos[ul] > y_pos[0]
    assert x_pos[lr] > x_pos[0] and y_pos[lr] < y_pos[0]
    assert x_pos[ur] > x_pos[0] and y_pos[ur] > y_pos[0]
    min_x = np.min(x_pos)
    min_y = np.min(y_pos)
    max_x = np.max(x_pos)
    max_y = np.max(y_pos)
    x_frac = (SCA_pos.x - min_x) / (max_x - min_x)
    y_frac = (SCA_pos.y - min_y) / (max_y - min_y)
    ll_ab = aberrations[ll, :]
    ul_ab = aberrations[ul, :]
    lr_ab = aberrations[lr, :]
    ur_ab = aberrations[ur, :]
    interp_ab = (1.0-x_frac)*(1.0-y_frac)*ll_ab + (1.0-x_frac)*y_frac*ul_ab + \
        x_frac*(1.0-y_frac)*lr_ab + x_frac*y_frac*ur_ab

    return interp_ab.flatten()



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'AstropyWCS', 'PyAstWCS', 'WcsToolsWCS', 'GSFitsWCS',
            'FitsWCS', 'TanWCS', 'FittedSIPWCS', ]

import warnings
import os
import numpy as np
import astropy.wcs
import subprocess
import copy

from .wcs import CelestialWCS, JacobianWCS, AffineTransform, PixelScale, OffsetWCS
from .position import PositionD, _PositionD
from .angle import radians, arcsec, degrees, AngleUnit
from . import _galsim
from . import fits
from .errors import GalSimError, GalSimValueError, GalSimIncompatibleValuesError
from .errors import GalSimNotImplementedError, convert_cpp_errors, galsim_warn
from .utilities import horner2d, least_squares
from .celestial import CelestialCoord
from ._pyfits import pyfits


#########################################################################################
#
# We have the following WCS classes that know how to read the WCS from a FITS file:
#
#     AstropyWCS
#     PyAstWCS
#     WcsToolsWCS
#     GSFitsWCS
#
# As for all CelestialWCS classes, they must define the following:
#
#     _radec            function returning (ra, dec) in _radians_ at position (x,y)
#     _xy               function returning (x, y) given (ra, dec) in _radians_.
#     _writeHeader      function that writes the WCS to a fits header.
#     _readHeader       static function that reads the WCS from a fits header.
#     copy              return a copy
#     __eq__            check if this equals another WCS
#
#########################################################################################


class AstropyWCS(CelestialWCS):
    """This WCS uses astropy.wcs to read WCS information from a FITS file.
    It requires the astropy.wcs python module to be installed.

    Astropy may be installed using pip, fink, or port::

        >>> pip install astropy
        >>> fink install astropy-py27
        >>> port install py27-astropy

    It also comes by default with Enthought and Anaconda. For more information, see their website:

        http://www.astropy.org/

    An AstropyWCS is initialized with one of the following commands::

        >>> wcs = galsim.AstropyWCS(file_name=file_name)  # Open a file on disk
        >>> wcs = galsim.AstropyWCS(header=header)        # Use an existing pyfits header
        >>> wcs = galsim.AstropyWCS(wcs=wcs)              # Use an existing astropy.wcs.WCS instance

    Exactly one of the parameters ``file_name``, ``header`` or ``wcs`` is required.  Also, since
    the most common usage will probably be the first, you can also give a ``file_name`` without it
    being named::

        >>> wcs = galsim.AstropyWCS(file_name)

    Parameters:
        file_name:        The FITS file from which to read the WCS information.  This is probably
                          the usual parameter to provide.  [default: None]
        dir:              Optional directory to prepend to ``file_name``. [default: None]
        hdu:              Optionally, the number of the HDU to use if reading from a file.
                          The default is to use either the primary or first extension as
                          appropriate for the given compression.  (e.g. for rice, the first
                          extension is the one you normally want.) [default: None]
        header:           The header of an open pyfits (or astropy.io) hdu.  Or, it can be
                          a FitsHeader object.  [default: None]
        compression:      Which decompression scheme to use (if any). See galsim.fits.read()
                          for the available options.  [default: 'auto']
        wcs:              An existing astropy.wcs.WCS instance [default: None]
        origin:           Optional origin position for the image coordinate system.
                          If provided, it should be a PositionD or PositionI.
                          [default: None]
    """
    _req_params = { "file_name" : str }
    _opt_params = { "dir" : str, "hdu" : int, "origin" : PositionD,
                    "compression" : str }

    def __init__(self, file_name=None, dir=None, hdu=None, header=None, compression='auto',
                 wcs=None, origin=None):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore",category=RuntimeWarning)
            import scipy  # AstropyWCS constructor will do this, so check now.
            import scipy.optimize # Check this too, since it's actually what we need from scipy.

        self._color = None
        self._tag = None # Write something useful here (see below). This is just used for the repr.
        self._set_origin(origin)

        # Read the file if given.
        if file_name is not None:
            if dir is not None:
                self._tag = repr(os.path.join(dir,file_name))
            else:
                self._tag = repr(file_name)
            if hdu is not None:
                self._tag += ', hdu=%r'%hdu
            if compression != 'auto':
                self._tag += ', compression=%r'%compression
            if header is not None:
                raise GalSimIncompatibleValuesError(
                    "Cannot provide both file_name and pyfits header",
                    file_name=file_name, header=header)
            if wcs is not None:
                raise GalSimIncompatibleValuesError(
                    "Cannot provide both file_name and wcs", file_name=file_name, wcs=wcs)
            hdu, hdu_list, fin = fits.readFile(file_name, dir, hdu, compression)

        try:
            if file_name is not None:
                header = hdu.header

            # Load the wcs from the header.
            if header is not None:
                if wcs is not None:
                    raise GalSimIncompatibleValuesError(
                        "Cannot provide both pyfits header and wcs", header=header, wcs=wcs)

                # These can mess things up later if they stick around.
                header.pop('BZERO', None)
                header.pop('BSCALE', None)

                self.header = fits.FitsHeader(header)
                try:
                    wcs = self._load_from_header(self.header)
                except Exception as e:  # pragma: no cover
                    # Not sure if this can still trigger.  There used to be input files that
                    # caused various errors in astropy, but that no longer seems to be true
                    # with astropy 4.x.  Leave this check here though, so the user can potentially
                    # get a more comprehensible error message if astropy fails.
                    raise OSError("Astropy failed to read WCS from %s. Original error: %s"%(
                                  file_name, e))
                else:
                    # New kind of error starting in astropy 2.0.5 (I think).  Sometimes, it
                    # gets through the above, but doesn't actually load the right WCS.
                    # E.g. ZPX gets marked as just a ZPN.
                    # As of version 4.0, ZPX is now the only one known to not work.  TPV has
                    # a similar behavior, but we can make it work by an adjustment to the header
                    # in _load_from_header.
                    if 'ZPX' in header.get('CTYPE1','') and 'ZPX' not in wcs.wcs.ctype[0]:
                        raise OSError(
                            "Cannot read WCS in %s with astropy. "%(file_name) +
                            "As of astropy version 4.0.1, ZPX WCS's were still not being " +
                            "correctly read by astropy.wcs. If you believe this has been " +
                            "fixed, please open a GalSim issue to remove this check.")
            else:
                self.header = None

            if wcs is None:
                raise GalSimIncompatibleValuesError(
                    "Must provide one of file_name, header, or wcs",
                    file_name=file_name, header=header, wcs=wcs)

        finally:
            if file_name is not None:
                fits.closeHDUList(hdu_list, fin)

        if not wcs.is_celestial:
            raise GalSimError("The WCS read in does not define a pair of celestial axes" )
        self._wcs = wcs

    def _load_from_header(self, header):
        if 'TAN' in header.get('CTYPE1','') and 'PV1_1' in header:
            header['CTYPE1'] = header['CTYPE1'].replace('TAN','TPV')
            header['CTYPE2'] = header['CTYPE2'].replace('TAN','TPV')
        with warnings.catch_warnings():
            # The constructor might emit warnings if it wants to fix the header
            # information (e.g. RADECSYS -> RADESYSa).  We'd rather ignore these
            # warnings, since we don't much care if the input file is non-standard
            # so long as we can make it work.
            warnings.simplefilter("ignore")
            wcs = astropy.wcs.WCS(header.header)
        return wcs

    @property
    def wcs(self):
        """The underlying ``astropy.wcs.WCS`` object.
        """
        return self._wcs

    @property
    def origin(self):
        """The origin in image coordinates of the WCS function.
        """
        return self._origin

    def _radec(self, x, y, color=None):
        x1 = np.atleast_1d(x)
        y1 = np.atleast_1d(y)

        ra, dec = self.wcs.all_pix2world(x1, y1, 1, ra_dec_order=True)

        # astropy outputs ra, dec in degrees.  Need to convert to radians.
        factor = degrees / radians

        if np.ndim(x) == np.ndim(y) == 0:
            return ra[0] * factor, dec[0] * factor
        else:
            # Sanity checks that the inputs are the same shape.
            assert np.ndim(x) == np.ndim(y)
            assert x.shape == y.shape
            ra *= factor
            dec *= factor
            return ra, dec

    def _xy(self, ra, dec, color=None):
        factor = radians / degrees

        r1 = np.atleast_1d(ra) * factor
        d1 = np.atleast_1d(dec) * factor

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            x, y = self.wcs.all_world2pix(r1, d1, 1, ra_dec_order=True)

        if np.ndim(ra) == np.ndim(dec) == 0:
            return x[0], y[0]
        else:
            # Sanity checks that the inputs are the same shape.
            assert np.ndim(ra) == np.ndim(dec)
            assert ra.shape == dec.shape
            return x, y

    def _newOrigin(self, origin):
        ret = self.copy()
        ret._origin = origin
        return ret

    def _writeHeader(self, header, bounds):
        # Make a new header with the contents of this WCS.
        # Note: relax = True means to write out non-standard FITS types.
        # Weirdly, this is the default when reading the header, but not when writing.
        header.update(self.wcs.to_header(relax=True))

        # And write the name as a special GalSim key
        header["GS_WCS"] = ("AstropyWCS", "GalSim WCS name")
        # And the image origin.
        header["GS_X0"] = (self.origin.x, "GalSim image origin x")
        header["GS_Y0"] = (self.origin.y, "GalSim image origin y")
        return header

    @staticmethod
    def _readHeader(header):
        x0 = header.get("GS_X0",0.)
        y0 = header.get("GS_Y0",0.)
        return AstropyWCS(header=header, origin=_PositionD(x0,y0))

    def copy(self):
        ret = AstropyWCS.__new__(AstropyWCS)
        ret.__dict__.update(self.__dict__)
        return ret

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, AstropyWCS) and
                 self.wcs.to_header(relax=True) == other.wcs.to_header(relax=True) and
                 self.origin == other.origin))

    def __repr__(self):
        if self._tag is not None:
            tag = self._tag
        elif self.header is not None:
            tag = 'header=%r'%self.header
        else:
            tag = 'wcs=%r'%self.wcs
        return "galsim.AstropyWCS(%s, origin=%r)"%(tag, self.origin)

    def __hash__(self): return hash(repr(self))

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_wcs']
        return d

    def __setstate__(self, d):
        self.__dict__ = d
        self._wcs = self._load_from_header(self.header)


class PyAstWCS(CelestialWCS): # pragma: no cover
    """This WCS uses PyAst (the python front end for the Starlink AST code) to read WCS
    information from a FITS file.  It requires the starlink.Ast python module to be installed.

    Starlink may be installed using pip::

        >>> pip install starlink-pyast

    For more information, see their website:

    https://pypi.python.org/pypi/starlink-pyast/

    A PyAstWCS is initialized with one of the following commands::

        >>> wcs = galsim.PyAstWCS(file_name=file_name)  # Open a file on disk
        >>> wcs = galsim.PyAstWCS(header=header)        # Use an existing pyfits header
        >>> wcs = galsim.PyAstWCS(wcsinfo=wcsinfo)      # Use an existing starlink.Ast.FrameSet

    Exactly one of the parameters ``file_name``, ``header`` or ``wcsinfo`` is required.  Also,
    since the most common usage will probably be the first, you can also give a file name without
    it being named::

        >>> wcs = galsim.PyAstWCS(file_name)

    Parameters:
        file_name:        The FITS file from which to read the WCS information.  This is probably
                          the usual parameter to provide.  [default: None]
        dir:              Optional directory to prepend to ``file_name``. [default: None]
        hdu:              Optionally, the number of the HDU to use if reading from a file.
                          The default is to use either the primary or first extension as
                          appropriate for the given compression.  (e.g. for rice, the first
                          extension is the one you normally want.) [default: None]
        header:           The header of an open pyfits (or astropy.io) hdu.  Or, it can be
                          a FitsHeader object.  [default: None]
        compression:      Which decompression scheme to use (if any). See galsim.fits.read()
                          for the available options.  [default:'auto']
        wcsinfo:          An existing starlink.Ast.FrameSet [default: None]
        origin:           Optional origin position for the image coordinate system.
                          If provided, it should be a PositionD or PositionI.
                          [default: None]
    """
    _req_params = { "file_name" : str }
    _opt_params = { "dir" : str, "hdu" : int, "origin" : PositionD,
                    "compression" : str }

    def __init__(self, file_name=None, dir=None, hdu=None, header=None, compression='auto',
                 wcsinfo=None, origin=None):
        self._color = None
        self._tag = None # Write something useful here (see below). This is just used for the repr.
        self._set_origin(origin)

        # Read the file if given.
        if file_name is not None:
            if dir is not None:
                self._tag = repr(os.path.join(dir,file_name))
            else:
                self._tag = repr(file_name)
            if hdu is not None:
                self._tag += ', hdu=%r'%hdu
            if compression != 'auto':
                self._tag += ', compression=%r'%compression
            if header is not None:
                raise GalSimIncompatibleValuesError(
                    "Cannot provide both file_name and pyfits header",
                    file_name=file_name, header=header)
            if wcsinfo is not None:
                raise GalSimIncompatibleValuesError(
                    "Cannot provide both file_name and wcsinfo",
                    file_name=file_name, wcsinfo=wcsinfo)
            hdu, hdu_list, fin = fits.readFile(file_name, dir, hdu, compression)

        try:
            if file_name is not None:
                header = hdu.header

            # Load the wcs from the header.
            if header is not None:
                if wcsinfo is not None:
                    raise GalSimIncompatibleValuesError(
                        "Cannot provide both pyfits header and wcsinfo",
                        header=header, wcsinfo=wcsinfo)

                # These can mess things up later if they stick around.
                header.pop('BZERO', None)
                header.pop('BSCALE', None)

                self.header = fits.FitsHeader(header)
                wcsinfo = self._load_from_header(self.header)
            else:
                self.header = None

            if wcsinfo is None:
                raise GalSimIncompatibleValuesError(
                    "Must provide one of file_name, header, or wcsinfo",
                    file_name=file_name, header=header, wcsinfo=wcsinfo)

            #  We can only handle WCS with 2 pixel axes (given by Nin) and 2 WCS axes
            # (given by Nout).
            if wcsinfo.Nin != 2 or wcsinfo.Nout != 2:  # pragma: no cover
                raise GalSimError("The world coordinate system is not 2-dimensional")

        finally:
            if file_name is not None:
                fits.closeHDUList(hdu_list, fin)

        self._wcsinfo = wcsinfo

    def _load_from_header(self, header):
        import starlink.Atl
        # Note: For much of this class implementation, I've followed the example provided here:
        #       http://dsberry.github.io/starlink/node4.html
        self._fix_header(header)

        # PyFITSAdapter requires an hdu, not a header, so just put it in a pyfits header object.
        # It turns out there are subtle differences between this and using the original FITS
        # file hdu that we read in above.  So there is a slight inefficiency here in creating
        # a new blank PrimaryHDU for this.  But in return we gain more reliable serializability.
        hdu = pyfits.PrimaryHDU()
        fits.FitsHeader(hdu_list=hdu).update(header)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # They aren't so good at keeping up with the latest pyfits and numpy syntax, so
            # this next line can emit deprecation warnings.
            # We can safely ignore them (for now...)
            fc = starlink.Ast.FitsChan(starlink.Atl.PyFITSAdapter(hdu))
            #  Read a FrameSet from the FITS header.
            wcsinfo = fc.read()

        if wcsinfo is None:
            raise OSError("Failed to read WCS information from fits file")

        # The PyAst WCS might not have (RA,Dec) axes, which we want.  It might for instance have
        # (Dec, RA) instead.  If it's possible to convert to an (RA,Dec) system, this next line
        # will do so.  And if not, the result will be None.
        # cf. https://github.com/timj/starlink-pyast/issues/8
        wcsinfo = wcsinfo.findframe(starlink.Ast.SkyFrame())
        if wcsinfo is None:
            raise GalSimError("The WCS read in does not define a pair of celestial axes" )

        return wcsinfo

    @property
    def wcsinfo(self):
        """The underlying ``starlink.Ast.FrameSet`` for this object.
        """
        return self._wcsinfo

    @property
    def origin(self):
        """The origin in image coordinates of the WCS function.
        """
        return self._origin

    def _fix_header(self, header):
        # We allow for the option to fix up the header information when a modification can
        # make it readable by PyAst.

        # There was an older proposed standard that used TAN with PV values, which is used by
        # SCamp, so we want to support it if possible.  The standard is now called TPV, which
        # PyAst understands.  All we need to do is change the names of the CTYPE values.
        if ( 'CTYPE1' in header and header['CTYPE1'].endswith('TAN') and
             'CTYPE2' in header and header['CTYPE2'].endswith('TAN') and
             'PV1_1' in header ):
            header['CTYPE1'] = header['CTYPE1'].replace('TAN','TPV')
            header['CTYPE2'] = header['CTYPE2'].replace('TAN','TPV')

    def _radec(self, x, y, color=None):
        # Need this to look like
        #    [ [ x1, x2, x3... ], [ y1, y2, y3... ] ]
        # if input is either scalar x,y or two arrays.
        xy = np.array([np.atleast_1d(x), np.atleast_1d(y)], dtype=float)

        ra, dec = self.wcsinfo.tran( xy )
        # PyAst returns ra, dec in radians, so we're good.

        if np.ndim(x) == np.ndim(y) == 0:
            return ra[0], dec[0]
        else:
            # Sanity checks that the inputs are the same shape.
            assert np.ndim(x) == np.ndim(y)
            assert x.shape == y.shape
            return ra, dec

    def _xy(self, ra, dec, color=None):
        rd = np.array([np.atleast_1d(ra), np.atleast_1d(dec)], dtype=float)
        x, y = self.wcsinfo.tran( rd, False )

        if np.ndim(ra) == np.ndim(dec) == 0:
            return x[0], y[0]
        else:
            # Sanity checks that the inputs are the same shape.
            assert np.ndim(ra) == np.ndim(dec)
            assert ra.shape == dec.shape
            return x, y

    def _newOrigin(self, origin):
        ret = self.copy()
        ret._origin = origin
        return ret

    def _writeHeader(self, header, bounds):
        import starlink.Atl
        # See https://github.com/Starlink/starlink/issues/24 for helpful information from
        # David Berry, who assisted me in getting this working.

        hdu = pyfits.PrimaryHDU()
        with warnings.catch_warnings():
            # Again, we can get deprecation warnings here.  Safe to ignore.
            warnings.simplefilter("ignore")
            fc = starlink.Ast.FitsChan(None, starlink.Atl.PyFITSAdapter(hdu) , "Encoding=FITS-WCS")
            # Let Ast know how big the image is that we'll be writing.
            for key in ('NAXIS', 'NAXIS1', 'NAXIS2'):
                if key in header:  # pragma: no branch
                    fc[key] = header[key]
            success = fc.write(self.wcsinfo)
            # PyAst doesn't write out TPV or ZPX correctly.  It writes them as TAN and ZPN
            # respectively.  However, if the maximum error is less than 0.1 pixel, it claims
            # success nonetheless.  This doesn't seem accurate enough for many purposes,
            # so we need to countermand that.
            # The easiest way I found to check for them is that the string TPN is in the string
            # version of wcsinfo.  So check for that and set success = False in that case.
            if 'TPN' in str(self.wcsinfo): success = False
            # Likewise for SIP.  MPF seems to be an appropriate string to look for.
            if 'MPF' in str(self.wcsinfo): success = False
            if not success:
                # This should always work, since it uses starlinks own proprietary encoding, but
                # it won't necessarily be readable by ds9.
                fc = starlink.Ast.FitsChan(None, starlink.Atl.PyFITSAdapter(hdu))
                fc.write(self.wcsinfo)
            fc.writefits()
        header.update(hdu.header)

        # And write the name as a special GalSim key
        header["GS_WCS"] = ("PyAstWCS", "GalSim WCS name")
        # And the image origin.
        header["GS_X0"] = (self.origin.x, "GalSim image origin x")
        header["GS_Y0"] = (self.origin.y, "GalSim image origin y")
        return header

    @staticmethod
    def _readHeader(header):
        x0 = header.get("GS_X0",0.)
        y0 = header.get("GS_Y0",0.)
        return PyAstWCS(header=header, origin=_PositionD(x0,y0))

    def copy(self):
        ret = PyAstWCS.__new__(PyAstWCS)
        ret.__dict__.update(self.__dict__)
        return ret

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, PyAstWCS) and
                 repr(self.wcsinfo) == repr(other.wcsinfo) and
                 self.origin == other.origin))

    def __repr__(self):
        if self._tag is not None:
            tag = self._tag
        elif self.header is not None:
            tag = 'header=%r'%self.header
        else:
            # Ast doesn't have a good repr for a FrameSet, so do it ourselves.
            tag = 'wcsinfo=<starlink.Ast.FrameSet at %s>'%id(self.wcsinfo)
        return "galsim.PyAstWCS(%s, origin=%r)"%(tag, self.origin)

    def __hash__(self): return hash(repr(self))

    def __getstate__(self):
        d = self.__dict__.copy()
        del d['_wcsinfo']
        return d

    def __setstate__(self, d):
        self.__dict__ = d
        self._wcsinfo = self._load_from_header(self.header)


# I can't figure out how to get wcstools installed in the travis environment (cf. .travis.yml).
# So until that gets resolved, we omit this class from the coverage report.
# This class was mostly useful as a refernce implementation anyway. It's much too slow for most
# users to ever want to use it.
class WcsToolsWCS(CelestialWCS): # pragma: no cover
    """This WCS uses wcstools executables to perform the appropriate WCS transformations
    for a given FITS file.  It requires wcstools command line functions to be installed.

    Note: It uses the wcstools executables xy2sky and sky2xy, so it can be quite a bit less
    efficient than other options that keep the WCS in memory.

    See their website for information on downloading and installing wcstools:

        http://tdc-www.harvard.edu/software/wcstools/

    A WcsToolsWCS is initialized with the following command::

        >>> wcs = galsim.WcsToolsWCS(file_name)

    Parameters:
        file_name:        The FITS file from which to read the WCS information.
        dir:              Optional directory to prepend to ``file_name``. [default: None]
        origin:           Optional origin position for the image coordinate system.
                          If provided, it should be a PositionD or PositionI.
                          [default: None]
    """
    _req_params = { "file_name" : str }
    _opt_params = { "dir" : str, "origin" : PositionD }

    def __init__(self, file_name, dir=None, origin=None):
        self._color = None
        self._set_origin(origin)

        if dir:
            file_name = os.path.join(dir, file_name)
        if not os.path.isfile(file_name):
            raise OSError('Cannot find file '+file_name)
        self._file_name = file_name

        # Check wcstools is installed and that it can read the file.
        # If xy2sky is not installed, this will raise an OSError
        p = subprocess.Popen(['xy2sky', '-d', '-n', '10', file_name, '0', '0'],
                             stdout=subprocess.PIPE)
        results = p.communicate()[0].decode()
        p.stdout.close()
        if len(results) == 0 or 'cannot' in results:
            raise OSError('wcstools (specifically xy2sky) was unable to read '+file_name)

        # wcstools supports LINEAR WCS's, but we don't want to allow them, since then
        # the CelestialWCS base class is inappropriate.  The clue to detect this is that
        # the results only have 4 values, rather than use usual 5 (missing epoch).
        if len(results.split()) == 4:
            raise GalSimError("The WCS read in does not define a pair of celestial axes" )

    @property
    def file_name(self):
        """The file name of the FITS file with the WCS information.
        """
        return self._file_name

    @property
    def origin(self):
        """The origin in image coordinates of the WCS function.
        """
        return self._origin

    def _radec(self, x, y, color=None):
        # Need this to look like
        #    [ x1, y1, x2, y2, ... ]
        # if input is either scalar x,y or two arrays.
        xy = np.array([x, y], dtype=float).transpose().ravel()

        # The OS cannot handle arbitrarily long command lines, so we may need to split up
        # the list into smaller chunks.
        if 'SC_ARG_MAX' in os.sysconf_names:
            arg_max = os.sysconf('SC_ARG_MAX')
        else:
            # A conservative guess. My machines have 131072, 262144, and 2621440
            arg_max = 32768

        # Sometimes SC_ARG_MAX is listed as -1.  Apparently that means "the configuration name
        # is known, but the value is not defined." So, just go with the above conservative value.
        if arg_max <= 0:
            arg_max = 32768

        # Just in case something weird happened.  This should be _very_ conservative.
        # It's the smallest value in this list of values for a bunch of systems:
        # http://www.in-ulm.de/~mascheck/various/argmax/
        if arg_max < 4096:
            arg_max = 4096

        # This corresponds to the total number of characters in the line.
        # But we really need to know how many arguments we are allowed to use in each call.
        # Lets be conservative again and assume each argument is at most 20 characters.
        # (We ignore the few characters at the start for the command name and such.)
        nargs = int(arg_max / 40) * 2  # Make sure it is even!

        xy_strs = [ str(z) for z in xy ]
        ra = []
        dec = []

        for i in range(0,len(xy_strs),nargs):
            xy1 = xy_strs[i:i+nargs]
            # We'd like to get the output to 10 digits of accuracy.  This corresponds to
            # an accuracy of about 1.e-6 arcsec.  But sometimes xy2sky cannot handle it,
            # in which case the output will start with *************.  If this happens, just
            # decrease digits and try again.
            for digits in range(10,5,-1):
                # If xy2sky is not installed, this will raise an OSError
                p = subprocess.Popen(['xy2sky', '-d', '-n', str(digits), self._file_name] + xy1,
                                    stdout=subprocess.PIPE)
                results = p.communicate()[0].decode()
                p.stdout.close()
                if len(results) == 0:
                    raise OSError('wcstools command xy2sky was unable to read '+ self._file_name)
                if results[0] != '*': break
            if results[0] == '*':
                raise OSError('wcstools command xy2sky was unable to read '+self._file_name)
            lines = results.splitlines()

            # Each line of output should looke like:
            #    x y J2000 ra dec
            # But if there was an error, the J200 might be missing or the output might look like
            #    Off map x y
            for line in lines:
                vals = line.split()
                if len(vals) != 5:
                    raise GalSimError('wcstools xy2sky returned invalid result near %s'%(xy1))
                ra.append(float(vals[0]))
                dec.append(float(vals[1]))

        # wcstools reports ra, dec in degrees, so convert to radians
        factor = degrees / radians

        if np.ndim(x) == np.ndim(y) == 0:
            return ra[0]*factor, dec[0]*factor
        else:
            # Sanity checks that the inputs are the same shape.
            assert np.ndim(x) == np.ndim(y)
            assert x.shape == y.shape
            return np.array(ra)*factor, np.array(dec)*factor

    def _xy(self, ra, dec, color=None):
        rd = np.array([ra, dec], dtype=float).transpose().ravel()
        rd *= radians / degrees

        # The boilerplate here is exactly the same as in _radec.  See that function for an
        # explanation of how this works.
        if 'SC_ARG_MAX' in os.sysconf_names:
            arg_max = os.sysconf('SC_ARG_MAX')
        else:
            arg_max = 32768
        if arg_max <= 0: arg_max = 32768
        if arg_max < 4096: arg_max = 4096
        nargs = int(arg_max / 40) * 2

        rd_strs = [ str(z) for z in rd ]
        x = []
        y = []

        for i in range(0,len(rd_strs),nargs):
            rd1 = rd_strs[i:i+nargs]
            for digits in range(10,5,-1):
                p = subprocess.Popen(['sky2xy', '-n', str(digits), self._file_name] + rd1,
                                    stdout=subprocess.PIPE)
                results = p.communicate()[0].decode()
                p.stdout.close()
                if len(results) == 0:
                    raise OSError('wcstools command sky2xy was unable to read '+self._file_name)
                if results[0] != '*': break
            if results[0] == '*':
                raise OSError('wcstools command sky2xy was unable to read '+self._file_name)

            lines = results.splitlines()

            # Each line of output should looke like:
            #    ra dec J2000 -> x y
            # However, if there was an error, the J200 might be missing.
            for line in lines:
                vals = line.split()
                if len(vals) < 6:
                    raise GalSimError('wcstools sky2xy returned invalid result for %f,%f'%(ra,dec))
                if len(vals) > 6:
                    galsim_warn("wcstools sky2xy indicates that %f,%f is off the image. "
                                "output is %r"%(ra,dec,results))
                x.append(float(vals[4]))
                y.append(float(vals[5]))

        if np.ndim(ra) == np.ndim(dec) == 0:
            return x[0], y[0]
        else:
            # Sanity checks that the inputs are the same shape.
            assert np.ndim(ra) == np.ndim(dec)
            assert ra.shape == dec.shape
            return np.array(x), np.array(y)

    def _newOrigin(self, origin):
        ret = self.copy()
        ret._origin = origin
        return ret

    def _writeHeader(self, header, bounds):
        # These are all we need to load it back.  Just use the original file.
        header["GS_WCS"]  = ("WcsToolsWCS", "GalSim WCS name")
        header["GS_FILE"] = (self._file_name, "GalSim original file with WCS data")
        header["GS_X0"] = (self.origin.x, "GalSim image origin x")
        header["GS_Y0"] = (self.origin.y, "GalSim image origin y")

        # We also copy over some of the fields we need.  wcstools doesn't seem to have something
        # that lists _all_ the keys that define the WCS.  This just gets the approximate WCS.
        p = subprocess.Popen(['wcshead', self._file_name], stdout=subprocess.PIPE)
        results = p.communicate()[0].decode()
        p.stdout.close()
        v = results.split()
        header["CTYPE1"] = v[3]
        header["CTYPE2"] = v[4]
        header["CRVAL1"] = v[5]
        header["CRVAL2"] = v[6]
        header["CRPIX1"] = v[8]
        header["CRPIX2"] = v[9]
        header["CDELT1"] = v[10]
        header["CDELT2"] = v[11]
        header["CROTA2"] = v[12]
        return header

    @staticmethod
    def _readHeader(header):
        file = header["GS_FILE"]
        x0 = header["GS_X0"]
        y0 = header["GS_Y0"]
        return WcsToolsWCS(file, origin=_PositionD(x0,y0))

    def copy(self):
        # The copy module version of copying the dict works fine here.
        return copy.copy(self)

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, WcsToolsWCS) and
                 self._file_name == other.file_name and
                 self.origin == other.origin))

    def __repr__(self):
        return "galsim.WcsToolsWCS(%r, origin=%r)"%(self._file_name, self.origin)

    def __hash__(self): return hash(repr(self))


class GSFitsWCS(CelestialWCS):
    """This WCS uses a GalSim implementation to read a WCS from a FITS file.

    It doesn't do nearly as many WCS types as the other options, and it does not try to be
    as rigorous about supporting all possible valid variations in the FITS parameters.
    However, it does several popular WCS types properly, and it doesn't require any additional
    python modules to be installed, which can be helpful.

    Currrently, it is able to parse the following WCS types: TAN, STG, ZEA, ARC, TPV, TNX

    A GSFitsWCS is initialized with one of the following commands::

        >>> wcs = galsim.GSFitsWCS(file_name=file_name)  # Open a file on disk
        >>> wcs = galsim.GSFitsWCS(header=header)        # Use an existing pyfits header

    Also, since the most common usage will probably be the first, you can also give a file name
    without it being named::

        >>> wcs = galsim.GSFitsWCS(file_name)

    In addition to reading from a FITS file, there is also a factory function that builds
    a GSFitsWCS object implementing a TAN projection.  See the docstring of `TanWCS` for
    more details.

    Parameters:
        file_name:        The FITS file from which to read the WCS information.  This is probably
                          the usual parameter to provide.  [default: None]
        dir:              Optional directory to prepend to ``file_name``. [default: None]
        hdu:              Optionally, the number of the HDU to use if reading from a file.
                          The default is to use either the primary or first extension as
                          appropriate for the given compression.  (e.g. for rice, the first
                          extension is the one you normally want.) [default: None]
        header:           The header of an open pyfits (or astropy.io) hdu.  Or, it can be
                          a FitsHeader object.  [default: None]
        compression:      Which decompression scheme to use (if any). See galsim.fits.read()
                          for the available options.  [default: 'auto']
        origin:           Optional origin position for the image coordinate system.
                          If provided, it should be a PositionD or PositionI.
                          [default: None]
    """
    _req_params = { "file_name" : str }
    _opt_params = { "dir" : str, "hdu" : int, "origin" : PositionD,
                    "compression" : str }

    def __init__(self, file_name=None, dir=None, hdu=None, header=None, compression='auto',
                 origin=None, _data=None, _doiter=True):
        # Note: _data is not intended for end-user use.  It enables the equivalent of a
        #       private constructor of GSFitsWCS by the function TanWCS.  The details of its
        #       use are intentionally not documented above.

        self._color = None
        self._tag = None # Write something useful here (see below). This is just used for the str.
        self._doiter = _doiter

        # If _data is given, copy the data and we're done.
        if _data is not None:
            self.wcs_type = _data[0]
            self.crpix = _data[1]
            self.cd = _data[2]
            self.center = _data[3]
            self.pv = _data[4]
            self.ab = _data[5]
            self.abp = _data[6]
            if self.wcs_type in ('TAN', 'TPV', 'TNX', 'TAN-SIP'):
                self.projection = 'gnomonic'
            elif self.wcs_type in ('STG', 'STG-SIP'):
                self.projection = 'stereographic'
            elif self.wcs_type in ('ZEA', 'ZEA-SIP'):
                self.projection = 'lambert'
            elif self.wcs_type in ('ARC', 'ARC-SIP'):
                self.projection = 'postel'
            else:
                raise ValueError("Invalid wcs_type in _data")
            return

        # Read the file if given.
        if file_name is not None:
            if dir is not None:
                self._tag = repr(os.path.join(dir,file_name))
            else:
                self._tag = repr(file_name)
            if hdu is not None:
                self._tag += ', hdu=%r'%hdu
            if compression != 'auto':
                self._tag += ', compression=%r'%compression
            if header is not None:
                raise GalSimIncompatibleValuesError(
                    "Cannot provide both file_name and pyfits header",
                    file_name=file_name, header=header)
            hdu, hdu_list, fin = fits.readFile(file_name, dir, hdu, compression)

        try:
            if file_name is not None:
                header = hdu.header

            if header is None:
                raise GalSimIncompatibleValuesError(
                    "Must provide either file_name or header", file_name=file_name, header=header)

            # Read the wcs information from the header.
            self._read_header(header)

        finally:
            if file_name is not None:
                fits.closeHDUList(hdu_list, fin)

        if origin is not None:
            self.crpix += [ origin.x, origin.y ]

    # The origin is a required attribute/property, since it is used by some functions like
    # shiftOrigin to get the current origin value.  We don't use it in this class, though, so
    # just make origin a dummy property that returns 0,0.
    @property
    def origin(self):
        """The origin in image coordinates of the WCS function.
        """
        return _PositionD(0.,0.)

    def _read_header(self, header):
        # Start by reading the basic WCS stuff that most types have.
        ctype1 = header.get('CTYPE1','')
        ctype2 = header.get('CTYPE2','')
        if ctype1.startswith('DEC--') and ctype2.startswith('RA---'):
            flip = True
        elif ctype1.startswith('RA---') and ctype2.startswith('DEC--'):
            flip = False
        else:
            raise GalSimError(
                "GSFitsWCS only supports celestial coordinate systems."
                "Expecting CTYPE1,2 to start with RA--- and DEC--.  Got %s, %s"%(ctype1, ctype2))
        if ctype1[5:] != ctype2[5:]:  # pragma: no cover
            raise OSError("ctype1, ctype2 do not seem to agree on the WCS type")
        self.wcs_type = ctype1[5:]
        if self.wcs_type in ('TAN', 'TPV', 'TNX', 'TAN-SIP'):
            self.projection = 'gnomonic'
        elif self.wcs_type in ('STG', 'STG-SIP'):
            self.projection = 'stereographic'
        elif self.wcs_type in ('ZEA', 'ZEA-SIP'):
            self.projection = 'lambert'
        elif self.wcs_type in ('ARC', 'ARC-SIP'):
            self.projection = 'postel'
        else:
            raise GalSimValueError("GSFitsWCS cannot read files using given wcs_type.",
                                   self.wcs_type,
                                   ('TAN', 'TPV', 'TNX', 'TAN-SIP', 'STG', 'STG-SIP', 'ZEA',
                                    'ZEA-SIP', 'ARC', 'ARC-SIP'))
        crval1 = float(header['CRVAL1'])
        crval2 = float(header['CRVAL2'])
        crpix1 = float(header['CRPIX1'])
        crpix2 = float(header['CRPIX2'])
        if 'CD1_1' in header:
            cd11 = float(header['CD1_1'])
            cd12 = float(header['CD1_2'])
            cd21 = float(header['CD2_1'])
            cd22 = float(header['CD2_2'])
        elif 'CDELT1' in header:
            if 'PC1_1' in header:
                cd11 = float(header['PC1_1']) * float(header['CDELT1'])
                cd12 = float(header['PC1_2']) * float(header['CDELT1'])
                cd21 = float(header['PC2_1']) * float(header['CDELT2'])
                cd22 = float(header['PC2_2']) * float(header['CDELT2'])
            else:
                cd11 = float(header['CDELT1'])
                cd12 = 0.
                cd21 = 0.
                cd22 = float(header['CDELT2'])
        else:  # pragma: no cover  (all our test files have either CD or CDELT)
            cd11 = 1.
            cd12 = 0.
            cd21 = 0.
            cd22 = 1.

        # Usually the units are degrees, but make sure
        if 'CUNIT1' in header:
            cunit1 = header['CUNIT1']
            cunit2 = header['CUNIT2']
            ra_units = AngleUnit.from_name(cunit1)
            dec_units = AngleUnit.from_name(cunit2)
        else:
            ra_units = degrees
            dec_units = degrees

        if flip:
            crval1, crval2 = crval2, crval1
            ra_units, dec_units = dec_units, ra_units
            cd11, cd21 = cd21, cd11
            cd12, cd22 = cd22, cd12

        self.crpix = np.array( [ crpix1, crpix2 ] )
        self.cd = np.array( [ [ cd11, cd12 ],
                              [ cd21, cd22 ] ] )

        self.center = CelestialCoord(crval1 * ra_units, crval2 * dec_units)

        # There was an older proposed standard that used TAN with PV values, which is used by
        # SCamp, so we want to support it if possible.  The standard is now called TPV, so
        # use that for our wcs_type if we see the PV values with TAN.
        if self.wcs_type == 'TAN' and 'PV1_1' in header:
            self.wcs_type = 'TPV'

        self.pv = None
        self.ab = None
        self.abp = None
        if self.wcs_type == 'TPV':
            self._read_tpv(header)
        elif self.wcs_type == 'TNX':
            self._read_tnx(header)
        elif self.wcs_type in ('TAN-SIP', 'STG-SIP', 'ZEA-SIP', 'ARC-SIP'):
            self._read_sip(header)

        # I think the CUNIT specification applies to the CD matrix as well, but I couldn't actually
        # find good documentation for this.  Plus all the examples I saw used degrees anyway, so
        # it's hard to tell.  Hopefully this will never matter, but if CUNIT is not deg, this
        # next bit might be wrong.
        # I did see documentation that the PV matrices always use degrees, so at least we shouldn't
        # have to worry about that.
        if ra_units != degrees:  # pragma: no cover
            self.cd[0,:] *= 1. * ra_units / degrees
        if dec_units != degrees:  # pragma: no cover
            self.cd[1,:] *= 1. * dec_units / degrees

    def _read_tpv(self, header):
        # See http://fits.gsfc.nasa.gov/registry/tpvwcs/tpv.html for details about how
        # the TPV standard is defined.

        # The standard includes an option to have odd powers of r, which kind of screws
        # up the numbering of these coefficients.  We don't implement these terms, so
        # before going further, check to make sure none are present.
        odd_indices = [3, 11, 23, 39]
        if any((header.get('PV%s_%s'%(i,j), 0.) != 0. for i in [1,2] for j in odd_indices)):
            raise GalSimNotImplementedError("TPV not implemented for odd powers of r")

        pv1 = [ float(header.get('PV1_%s'%k, 0.)) for k in range(40) if k not in odd_indices ]
        pv2 = [ float(header.get('PV2_%s'%k, 0.)) for k in range(40) if k not in odd_indices ]

        maxk = max(np.nonzero(pv1)[0][-1], np.nonzero(pv2)[0][-1])
        # maxk = (order+1) * (order+2) / 2 - 1
        order = int(np.floor(np.sqrt(2*(maxk+1)))) - 1
        self.pv = np.zeros((2,order+1,order+1))

        # Another strange thing is that the two matrices are defined in the opposite order
        # with respect to their element ordering.  But at least now, without the odd terms,
        # we can just proceed in order in the k indices.  So what we call k=3..9 here were
        # originally PVi_4..10.
        # For reference, here is what it would look like for order = 3:
        # self.pv = np.array( [ [ [ pv1[0], pv1[2], pv1[5], pv1[9] ],
        #                         [ pv1[1], pv1[4], pv1[8],   0.   ],
        #                         [ pv1[3], pv1[7],   0.  ,   0.   ],
        #                         [ pv1[6],   0.  ,   0.  ,   0.   ] ],
        #                       [ [ pv2[0], pv2[1], pv2[3], pv2[6] ],
        #                         [ pv2[2], pv2[4], pv2[7],   0.   ],
        #                         [ pv2[5], pv2[8],   0.  ,   0.   ],
        #                         [ pv2[9],   0.  ,   0.  ,   0.   ] ] ] )
        k = 0
        for N in range(order+1):
            for j in range(N+1):
                i = N-j
                self.pv[0,i,j] = pv1[k]
                self.pv[1,j,i] = pv2[k]
                k = k+1

    def _read_sip(self, header):
        a_order = int(header['A_ORDER'])
        b_order = int(header['B_ORDER'])
        order = max(a_order,b_order)  # Use the same order for both
        a = [ float(header.get('A_'+str(i)+'_'+str(j),0.))
                for i in range(order+1) for j in range(order+1) ]
        a = np.array(a).reshape((order+1,order+1))
        b = [ float(header.get('B_'+str(i)+'_'+str(j),0.))
                for i in range(order+1) for j in range(order+1) ]
        b = np.array(b).reshape((order+1,order+1))
        a[1,0] += 1  # Standard A,B are a differential calculation.  It's more convenient to
        b[0,1] += 1  # keep this as an absolute calculation like PV does.
        self.ab = np.array([a, b])

        # The reverse transformation is not required to be there.
        if 'AP_ORDER' in header:
            ap_order = int(header['AP_ORDER'])
            bp_order = int(header['BP_ORDER'])
            order = max(ap_order,bp_order)  # Use the same order for both
            ap = [ float(header.get('AP_'+str(i)+'_'+str(j),0.))
                    for i in range(order+1) for j in range(order+1) ]
            ap = np.array(ap).reshape((order+1,order+1))
            bp = [ float(header.get('BP_'+str(i)+'_'+str(j),0.))
                    for i in range(order+1) for j in range(order+1) ]
            bp = np.array(bp).reshape((order+1,order+1))
            ap[1,0] += 1
            bp[0,1] += 1
            self.abp = np.array([ap, bp])

    def _read_tnx(self, header):

        # TNX has a few different options.  Rather than keep things in the native format,
        # we actually convert to the equivalent of TPV to make the actual operations faster.
        # See http://iraf.noao.edu/projects/ccdmosaic/tnx.html for details.

        # First, parse the input values, which are stored in WAT keywords:
        k = 1
        wat1 = ""
        key = 'WAT1_%03d'%k
        while key in header:
            wat1 += header[key]
            k = k+1
            key = 'WAT1_%03d'%k
        wat1 = wat1.split()

        k = 1
        wat2 = ""
        key = 'WAT2_%03d'%k
        while key in header:
            wat2 += header[key]
            k = k+1
            key = 'WAT2_%03d'%k
        wat2 = wat2.split()

        if ( len(wat1) < 12 or
             wat1[0] != 'wtype=tnx' or
             wat1[1] != 'axtype=ra' or
             wat1[2] != 'lngcor' or
             wat1[3] != '=' or
             not wat1[4].startswith('"') or
             not wat1[-1].endswith('"') ):  # pragma: no cover
            raise GalSimError("TNX WAT1 was not as expected")
        if ( len(wat2) < 12 or
             wat2[0] != 'wtype=tnx' or
             wat2[1] != 'axtype=dec' or
             wat2[2] != 'latcor' or
             wat2[3] != '=' or
             not wat2[4].startswith('"') or
             not wat2[-1].endswith('"') ):  # pragma: no cover
            raise GalSimError("TNX WAT2 was not as expected")

        # Break the next bit out into another function, since it is the same for x and y.
        pv1 = self._parse_tnx_data(wat1[4:])
        pv2 = self._parse_tnx_data(wat2[4:])

        # Those just give the adjustments to the position, not the matrix that gives the final
        # position.  i.e. the TNX standard uses u = u + [1 u u^2 u^3] PV [1 v v^2 v^3]T.
        # So we need to add 1 to the correct term in each matrix to get what we really want.
        pv1[1,0] += 1.
        pv2[0,1] += 1.

        # Finally, store these as our pv 3-d array.
        self.pv = np.array([pv1, pv2])

        # We've now converted this to TPV, so call it that when we output to a fits header.
        self.wcs_type = 'TPV'

    def _parse_tnx_data(self, data):

        # I'm not sure if there is any requirement on there being a space before the final " and
        # not before the initial ".  But both the example in the description of the standard and
        # the one we have in our test directory are this way.  Here, if the " is by itself, I
        # remove the item, and if it is part of a longer string, I just strip it off.  Seems the
        # most sensible thing to do.
        if data[0] == '"':  # pragma: no cover
            data = data[1:]
        else:
            data[0] = data[0][1:]
        if data[-1] == '"':
            data = data[:-1]
        else:  # pragma: no cover
            data[-1] = data[-1][:-1]

        code = int(data[0].strip('.'))  # Weirdly, these integers are given with decimal points.
        xorder = int(data[1].strip('.'))
        yorder = int(data[2].strip('.'))
        cross = int(data[3].strip('.'))
        if cross != 2:  # pragma: no cover
            raise GalSimNotImplementedError("TNX only implemented for half-cross option.")
        if xorder != 4 or yorder != 4:  # pragma: no cover
            raise GalSimNotImplementedError("TNX only implemented for order = 4")
        # Note: order = 4 really means cubic.  order is how large the pv matrix is, i.e. 4x4.

        xmin = float(data[4])
        xmax = float(data[5])
        ymin = float(data[6])
        ymax = float(data[7])

        pv1 = [ float(x) for x in data[8:] ]
        if len(pv1) != 10:  # pragma: no cover
            raise GalSimError("Wrong number of items found in WAT data")

        # Put these into our matrix formulation.
        pv = np.array( [ [ pv1[0], pv1[4], pv1[7], pv1[9] ],
                         [ pv1[1], pv1[5], pv1[8],   0.   ],
                         [ pv1[2], pv1[6],   0.  ,   0.   ],
                         [ pv1[3],   0.  ,   0.  ,   0.   ] ] )

        # Convert from Legendre or Chebyshev polynomials into regular polynomials.
        if code < 3: # pragma: no branch (The only test file I can find has code = 1)
            # Instead of 1, x, x^2, x^3, Chebyshev uses: 1, x', 2x'^2 - 1, 4x'^3 - 3x
            # where x' = (2x - xmin - xmax) / (xmax-xmin).
            # Similarly, with y' = (2y - ymin - ymin) / (ymax-ymin)
            # We'd like to convert the pv matrix from being in terms of x' and y' to being
            # in terms of just x, y.  To see how this works, look at what pv[1,1] means:
            #
            # First, let's say we can write x as (a + bx), and we can write y' as (c + dy).
            # Then the term for pv[1,1] is:
            #
            # term = x' * pv[1,1] * y'
            #      = (a + bx) * pv[1,1] * (d + ey)
            #      =       a * pv[1,1] * c  +      a * pv[1,1] * d * y
            #        + x * b * pv[1,1] * c  +  x * b * pv[1,1] * d * y
            #
            # So the single term initially will contribute to 4 different terms in the final
            # matrix.  And the contributions will just be pv[1,1] times the outer product
            # [a b]T [d e].  So if we can determine the matrix that converts from
            # [1, x, x^2, x^3] to the Chebyshev vector, the the matrix we want is simply
            # xmT pv ym.
            a = -(xmax+xmin)/(xmax-xmin)
            b = 2./(xmax-xmin)
            c = -(ymax+ymin)/(ymax-ymin)
            d = 2./(ymax-ymin)
            xm = np.zeros((4,4))
            ym = np.zeros((4,4))
            xm[0,0] = 1.
            xm[1,0] = a
            xm[1,1] = b
            ym[0,0] = 1.
            ym[1,0] = c
            ym[1,1] = d
            if code == 1:
                for m in range(2,4):
                    # The recursion rule is Pm = 2 x' Pm-1 - Pm-2
                    # Pm = 2 a Pm-1 - Pm-2 + x * 2 b Pm-1
                    xm[m] = 2. * a * xm[m-1] - xm[m-2]
                    xm[m,1:] += 2. * b * xm[m-1,:-1]
                    ym[m] = 2. * c * ym[m-1] - ym[m-2]
                    ym[m,1:] += 2. * d * ym[m-1,:-1]
            else:  # pragma: no cover
                # code == 2 means Legendre.  The same argument applies, but we have a
                # different recursion rule.
                # WARNING: This branch has not been tested!  I don't have any TNX files
                # with Legendre functions to test it on.  I think it's right, but beware!
                for m in range(2,4):
                    # The recursion rule is Pm = ((2m-1) x' Pm-1 - (m-1) Pm-2) / m
                    # Pm = ((2m-1) a Pm-1 - (m-1) Pm-2) / m
                    #      + x * ((2m-1) b Pm-1) / m
                    xm[m] = ((2.*m-1.) * a * xm[m-1] - (m-1.) * xm[m-2]) / m
                    xm[m,1:] += ((2.*m-1.) * b * xm[m-1,:-1]) / m
                    ym[m] = ((2.*m-1.) * c * ym[m-1] - (m-1.) * ym[m-2]) / m
                    ym[m,1:] += ((2.*m-1.) * d * ym[m-1,:-1]) / m

            pv2 = np.dot(xm.T , np.dot(pv, ym))
            return pv2

    def _apply_ab(self, x, y, ab):
        # Note: this is used for both pv and ab, since the action is the same.
        # They just occur at two different places in the calculation.
        x1 = horner2d(x, y, ab[0], triangle=True)
        y1 = horner2d(x, y, ab[1], triangle=True)
        return x1, y1

    def _apply_cd(self, x, y):
        # Do this in C++ layer for speed.
        nx = len(x.ravel())
        _x = x.__array_interface__['data'][0]
        _y = y.__array_interface__['data'][0]
        _cd = self.cd.__array_interface__['data'][0]
        _galsim.ApplyCD(nx, _x, _y, _cd)
        return x, y

    def _uv(self, x, y):
        # Most of the work for _radec.  But stop at (u,v).

        # Start with (u,v) = the image position
        x = np.ascontiguousarray(x, dtype=float)
        y = np.ascontiguousarray(y, dtype=float)

        x -= self.crpix[0]
        y -= self.crpix[1]

        if self.ab is not None:
            x, y = self._apply_ab(x, y, self.ab)

        # This converts to (u,v) in the tangent plane
        # Expanding this out is a bit faster than using np.dot for 2x2 matrix.
        u, v = self._apply_cd(x, y)

        if self.pv is not None:
            u, v = self._apply_ab(u, v, self.pv)

        # Convert (u,v) from degrees to radians
        # Also, the FITS standard defines u,v backwards relative to our standard.
        # They have +u increasing to the east, not west.  Hence the - for u.
        factor = 1. * degrees / radians
        u *= -factor
        v *= factor
        return u, v

    def _radec(self, x, y, color=None):
        # Get the position in the tangent plane
        u, v = self._uv(x, y)
        # Then convert from (u,v) to (ra, dec) using the appropriate projection.
        ra, dec = self.center.deproject_rad(u, v, projection=self.projection)

        if np.ndim(x) == np.ndim(y) == 0:
            return ra[0], dec[0]
        else:
            # Sanity checks that the inputs are the same shape.
            assert np.ndim(x) == np.ndim(y)
            assert x.shape == y.shape
            return ra, dec

    def _invert_ab(self, u, v, ab, abp=None):
        # This is used both for inverting (u,v) = PV (u',v')
        # and for inverting (x,y) = AB (x',y')
        # Here (and in C++) the notation is (u,v) = AB(x,y), even though both (u,v) and (x,y)
        # in this context are either in CCD coordinates (normally called x,y) or tangent plane
        # coordinates (normally called u,v).
        # abp is an optional set of coefficients to make a good guess for x,y

        uu = np.ascontiguousarray(u)  # Don't overwrite the given u,v, since we need it at the end
        vv = np.ascontiguousarray(v)  # to check it we were provided scalars or arrays.

        x = np.atleast_1d(u.copy())    # Start with x,y = u,v.
        y = np.atleast_1d(v.copy())    # This may be updated below if abp is provided.

        nab = ab.shape[1]
        nabp = abp.shape[1] if abp is not None else 0
        nx = len(x.ravel())
        _uu = uu.__array_interface__['data'][0]
        _vv = vv.__array_interface__['data'][0]
        _x = x.__array_interface__['data'][0]
        _y = y.__array_interface__['data'][0]
        _ab = ab.__array_interface__['data'][0]
        _abp = 0 if abp is None else abp.__array_interface__['data'][0]
        with convert_cpp_errors():
            _galsim.InvertAB(nx, nab, _uu, _vv, _ab, _x, _y, self._doiter, nabp, _abp)

        # Return the right type for u,v
        try:
            len(u)
        except TypeError:
            return x[0], y[0]
        else:
            return x, y


    def _xy(self, ra, dec, color=None):
        u, v = self.center.project_rad(ra, dec, projection=self.projection)

        # Again, FITS has +u increasing to the east, not west.  Hence the - for u.
        factor = radians / degrees
        u *= -factor
        v *= factor

        if self.pv is not None:
            u, v = self._invert_ab(u, v, self.pv)

        if not hasattr(self, 'cdinv'):
            self.cdinv = np.linalg.inv(self.cd)
        # This is a bit faster than using np.dot for 2x2 matrix.
        x = self.cdinv[0,0] * u + self.cdinv[0,1] * v
        y = self.cdinv[1,0] * u + self.cdinv[1,1] * v

        if self.ab is not None:
            x, y = self._invert_ab(x, y, self.ab, abp=self.abp)

        x += self.crpix[0]
        y += self.crpix[1]

        return x, y

    # Override the version in CelestialWCS, since we can do this more efficiently.
    def _local(self, image_pos, color=None):

        if image_pos is None:
            raise TypeError("origin must be a PositionD or PositionI argument")

        # The key lemma here is that chain rule for jacobians is just matrix multiplication.
        # i.e. if s = s(u,v), t = t(u,v) and u = u(x,y), v = v(x,y), then
        # ( dsdx  dsdy ) = ( dsdu dudx + dsdv dvdx   dsdu dudy + dsdv dvdy )
        # ( dtdx  dtdy ) = ( dtdu dudx + dtdv dvdx   dtdu dudy + dtdv dvdy )
        #                = ( dsdu  dsdv )  ( dudx  dudy )
        #                  ( dtdu  dtdv )  ( dvdx  dvdy )
        #
        # So if we can find the jacobian for each step of the process, we just multiply the
        # jacobians.
        #
        # We also need to keep track of the position along the way, so we have to repeat many
        # of the steps in _radec.

        p1 = np.array([image_pos.x, image_pos.y], dtype=float)

        # Start with unit jacobian
        jac = np.diag([1,1])

        # No effect on the jacobian from this step.
        p1 -= self.crpix

        if self.ab is not None:
            x = p1[0]
            y = p1[1]
            order = len(self.ab[0])-1
            xpow = x ** np.arange(order+1)
            ypow = y ** np.arange(order+1)
            p1 = np.dot(np.dot(self.ab, ypow), xpow)

            dxpow = np.zeros(order+1)
            dypow = np.zeros(order+1)
            dxpow[1:] = (np.arange(order)+1.) * xpow[:-1]
            dypow[1:] = (np.arange(order)+1.) * ypow[:-1]
            j1 = np.transpose([ np.dot(np.dot(self.ab, ypow), dxpow) ,
                                np.dot(np.dot(self.ab, dypow), xpow) ])
            jac = np.dot(j1,jac)

        # The jacobian here is just the cd matrix.
        p2 = np.dot(self.cd, p1)
        jac = np.dot(self.cd, jac)

        if self.pv is not None:
            # Now we apply the distortion terms
            u = p2[0]
            v = p2[1]
            order = len(self.pv[0])-1

            upow = u ** np.arange(order+1)
            vpow = v ** np.arange(order+1)

            p2 = np.dot(np.dot(self.pv, vpow), upow)

            # The columns of the jacobian for this step are the same function with dupow
            # or dvpow.
            dupow = np.zeros(order+1)
            dvpow = np.zeros(order+1)
            dupow[1:] = (np.arange(order)+1.) * upow[:-1]
            dvpow[1:] = (np.arange(order)+1.) * vpow[:-1]
            j1 = np.transpose([ np.dot(np.dot(self.pv, vpow), dupow) ,
                                np.dot(np.dot(self.pv, dvpow), upow) ])
            jac = np.dot(j1,jac)

        unit_convert = [ -1 * degrees / radians, 1 * degrees / radians ]
        p2 *= unit_convert
        # Subtle point: Don't use jac *= ..., because jac might currently be self.cd, and
        #               that would change self.cd!
        jac = jac * np.transpose( [ unit_convert ] )

        # Finally convert from (u,v) to (ra, dec).  We have a special function that computes
        # the jacobian of this step in the CelestialCoord class.
        j2 = self.center.jac_deproject_rad(p2[0], p2[1], projection=self.projection)
        jac = np.dot(j2,jac)

        # This now has units of radians/pixel.  We want instead arcsec/pixel.
        jac *= radians / arcsec

        return JacobianWCS(jac[0,0], jac[0,1], jac[1,0], jac[1,1])


    def _newOrigin(self, origin):
        ret = self.copy()
        ret.crpix = ret.crpix + [ origin.x, origin.y ]
        return ret

    def _writeHeader(self, header, bounds):
        header["GS_WCS"] = ("GSFitsWCS", "GalSim WCS name")
        header["CTYPE1"] = 'RA---' + self.wcs_type
        header["CTYPE2"] = 'DEC--' + self.wcs_type
        header["CRPIX1"] = self.crpix[0]
        header["CRPIX2"] = self.crpix[1]
        header["CD1_1"] = self.cd[0][0]
        header["CD1_2"] = self.cd[0][1]
        header["CD2_1"] = self.cd[1][0]
        header["CD2_2"] = self.cd[1][1]
        header["CUNIT1"] = 'deg'
        header["CUNIT2"] = 'deg'
        header["CRVAL1"] = self.center.ra / degrees
        header["CRVAL2"] = self.center.dec / degrees
        if self.pv is not None:
            order = len(self.pv[0])-1
            k = 0
            odd_indices = [3, 11, 23, 39]
            for n in range(order+1):
                for j in range(n+1):
                    i = n-j
                    header["PV1_" + str(k)] = self.pv[0, i, j]
                    header["PV2_" + str(k)] = self.pv[1, j, i]
                    k = k + 1
                    if k in odd_indices: k = k + 1
        if self.ab is not None:
            order = len(self.ab[0])-1
            header["A_ORDER"] = order
            for i in range(order+1):
                for j in range(order+1):
                    aij = self.ab[0,i,j]
                    if i==1 and j==0: aij -= 1  # Turn back into standard form.
                    if aij != 0.:
                        header["A_"+str(i)+"_"+str(j)] = aij
            header["B_ORDER"] = order
            for i in range(order+1):
                for j in range(order+1):
                    bij = self.ab[1,i,j]
                    if i==0 and j==1: bij -= 1
                    if bij != 0.:
                        header["B_"+str(i)+"_"+str(j)] = bij
        if self.abp is not None:
            order = len(self.abp[0])-1
            header["AP_ORDER"] = order
            for i in range(order+1):
                for j in range(order+1):
                    apij = self.abp[0,i,j]
                    if i==1 and j==0: apij -= 1
                    if apij != 0.:
                        header["AP_"+str(i)+"_"+str(j)] = apij
            header["BP_ORDER"] = order
            for i in range(order+1):
                for j in range(order+1):
                    bpij = self.abp[1,i,j]
                    if i==0 and j==1: bpij -= 1
                    if bpij != 0.:
                        header["BP_"+str(i)+"_"+str(j)] = bpij
        return header

    @staticmethod
    def _readHeader(header):
        return GSFitsWCS(header=header)

    def copy(self):
        # The copy module version of copying the dict works fine here.
        return copy.copy(self)

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, GSFitsWCS) and
                 self.wcs_type == other.wcs_type and
                 np.array_equal(self.crpix,other.crpix) and
                 np.array_equal(self.cd,other.cd) and
                 self.center == other.center and
                 np.array_equal(self.pv,other.pv) and
                 np.array_equal(self.ab,other.ab) and
                 np.array_equal(self.abp,other.abp)))

    def __repr__(self):
        if self.pv is None:
            pv_repr = repr(self.pv)
        else:
            pv_repr = 'array(%r)'%self.pv.tolist()
        if self.ab is None:
            ab_repr = repr(self.ab)
        else:
            ab_repr = 'array(%r)'%self.ab.tolist()
        if self.abp is None:
            abp_repr = repr(self.abp)
        else:
            abp_repr = 'array(%r)'%self.abp.tolist()
        return "galsim.GSFitsWCS(_data = [%r, array(%r), array(%r), %r, %s, %s, %s])"%(
                self.wcs_type, self.crpix.tolist(), self.cd.tolist(), self.center,
                pv_repr, ab_repr, abp_repr)

    def __str__(self):
        if self._tag is None:
            return self.__repr__()
        else:
            return "galsim.GSFitsWCS(%s)"%(self._tag)

    def __hash__(self): return hash(repr(self))


def TanWCS(affine, world_origin, units=arcsec):
    """This is a function that returns a `GSFitsWCS` object for a TAN WCS projection.

    The TAN projection is essentially an affine transformation from image coordinates to
    Euclidean (u,v) coordinates on a tangent plane, and then a "deprojection" of this plane
    onto the sphere given a particular RA, Dec for the location of the tangent point.
    The tangent point will correspond to the location of (u,v) = (0,0) in the intermediate
    coordinate system.

    Parameters:
        affine:          An `AffineTransform` defining the transformation from image coordinates
                         to the coordinates on the tangent plane.
        world_origin:    A `CelestialCoord` defining the location on the sphere where the
                         tangent plane is centered.
        units:           The angular units of the (u,v) intermediate coordinate system.
                         [default: galsim.arcsec]

    Returns:
        a `GSFitsWCS` describing this WCS.
    """
    # These will raise the appropriate errors if affine is not the right type.
    dudx = affine.dudx * units / degrees
    dudy = affine.dudy * units / degrees
    dvdx = affine.dvdx * units / degrees
    dvdy = affine.dvdy * units / degrees
    origin = affine.origin
    # The - signs are because the Fits standard is in terms of +u going east, rather than west
    # as we have defined.  So just switch the sign in the CD matrix.
    cd = np.array([[ -dudx, -dudy ], [ dvdx, dvdy ]], dtype=float)
    crpix = np.array([ origin.x, origin.y ], dtype=float)

    # We also need to absorb the affine world_origin back into crpix, since GSFits is expecting
    # crpix to be the location of the tangent point in image coordinates. i.e. where (u,v) = (0,0)
    # (u,v) = CD * (x-x0,y-y0) + (u0,v0)
    # (0,0) = CD * (x0',y0') - CD * (x0,y0) + (u0,v0)
    # CD (x0',y0') = CD (x0,y0) - (u0,v0)
    # (x0',y0') = (x0,y0) - CD^-1 (u0,v0)
    uv = np.array( [ affine.world_origin.x * units / degrees,
                     affine.world_origin.y * units / degrees ] )
    crpix -= np.dot(np.linalg.inv(cd) , uv)

    # Invoke the private constructor of GSFits using the _data kwarg.
    data = ('TAN', crpix, cd, world_origin, None, None, None)
    return GSFitsWCS(_data=data)


# This is a list of all the WCS types that can potentially read a WCS from a FITS file.
# The function FitsWCS will try each of these in order and return the first one that
# succeeds.  AffineTransform should be last, since it will always succeed.
# The list is defined here at global scope so that external modules can add extra
# WCS types to the list if desired.

fits_wcs_types = [

    GSFitsWCS,      # This doesn't work for very many WCS types, but it works for the very common
                    # TAN projection, and also TPV, which is used by SCamp.  If it does work, it
                    # is a good choice, since it is easily the fastest of any of these.

    PyAstWCS,       # This requires ``import starlink.Ast`` to succeed.  This handles the largest
                    # number of WCS types of any of these.  In fact, it worked for every one
                    # we tried in our unit tests (which was not exhaustive).

    AstropyWCS,     # This requires ``import astropy.wcs`` to succeed.  It doesn't support quite as
                    # many WCS types as PyAst.  It's also usually a little slower, so we prefer
                    # PyAstWCS when it is available.

    WcsToolsWCS,    # This requires the wcstool command line functions to be installed.
                    # It is very slow, so it should only be used as a last resort.

]


def FitsWCS(file_name=None, dir=None, hdu=None, header=None, compression='auto',
            text_file=False, suppress_warning=False):
    """This factory function will try to read the WCS from a FITS file and return a WCS that will
    work.  It tries a number of different WCS classes until it finds one that succeeds in reading
    the file.

    If none of them work, then the last class it tries, `AffineTransform`, is guaranteed to succeed,
    but it will only model the linear portion of the WCS (the CD matrix, CRPIX, and CRVAL), using
    reasonable defaults if even these are missing.  If you think that you have the right software
    for one of the WCS types, but FitsWCS still defaults to `AffineTransform`, it may be helpful to
    update your installation of astropy and/or starlink (if you don't already have the latest
    version).

    Note: The list of classes this function will try may be edited, e.g. by an external module
    that wants to add an additional WCS type.  The list is ``galsim.fitswcs.fits_wcs_types``.

    Parameters:
        file_name:        The FITS file from which to read the WCS information.  This is probably
                          the usual parameter to provide.  [default: None]
        dir:              Optional directory to prepend to ``file_name``. [default: None]
        hdu:              Optionally, the number of the HDU to use if reading from a file.
                          The default is to use either the primary or first extension as
                          appropriate for the given compression.  (e.g. for rice, the first
                          extension is the one you normally want.) [default: None]
        header:           The header of an open pyfits (or astropy.io) hdu.  Or, it can be
                          a FitsHeader object.  [default: None]
        compression:      Which decompression scheme to use (if any). See galsim.fits.read()
                          for the available options.  [default: 'auto']
        text_file:        Normally a file is taken to be a fits file, but you can also give it a
                          text file with the header information (like the .head file output from
                          SCamp).  In this case you should set ``text_file = True`` to tell GalSim
                          to parse the file this way.  [default: False]
        suppress_warning: Whether to suppress a warning that the WCS could not be read from the
                          FITS header, so the WCS defaulted to either a `PixelScale` or
                          `AffineTransform`. [default: False]
                          (Note: this is (by default) set to True when this function is implicitly
                          called from one of the galsim.fits.read* functions.)
    """
    if file_name is not None:
        if header is not None:
            raise GalSimIncompatibleValuesError(
                "Cannot provide both file_name and pyfits header",
                file_name=file_name, header=header)
        header = fits.FitsHeader(file_name=file_name, dir=dir, hdu=hdu, compression=compression,
                                 text_file=text_file)
    else:
        file_name = 'header' # For sensible error messages below.
    if header is None:
        raise GalSimIncompatibleValuesError(
            "Must provide either file_name or header", file_name=file_name, header=header)
    if not isinstance(header, fits.FitsHeader):
        header = fits.FitsHeader(header)

    if 'CTYPE1' not in header and 'CDELT1' not in header:
        if not suppress_warning:
            galsim_warn("No WCS information found in %r. Defaulting to PixelScale(1.0)"%(file_name))
        return PixelScale(1.0)

    # For linear WCS specifications, AffineTransformation should work.
    # Note: Most files will have CTYPE1,2, but old style with only CDELT1,2 sometimes omits it.
    if header.get('CTYPE1', 'LINEAR') == 'LINEAR':
        wcs = AffineTransform._readHeader(header)
        # Convert to PixelScale if possible.
        if (wcs.dudx == wcs.dvdy and wcs.dudy == wcs.dvdx == 0):
            if wcs.x0 == wcs.y0 == wcs.u0 == wcs.v0 == 0:
                wcs = PixelScale(wcs.dudx)
            else:
                wcs = OffsetWCS(wcs.dudx, wcs.origin, wcs.world_origin)
        return wcs

    # Otherwise (and typically), try the various wcs types that can read celestial coordinates.
    for wcs_type in fits_wcs_types:
        try:
            wcs = wcs_type._readHeader(header)
            # Give it a better tag for the repr if appropriate.
            if hasattr(wcs,'_tag') and file_name != 'header':
                if dir is not None:
                    wcs._tag = repr(os.path.join(dir,file_name))
                else:
                    wcs._tag = repr(file_name)
                if hdu is not None:
                    wcs._tag += ', hdu=%r'%hdu
                if compression != 'auto':
                    wcs._tag += ', compression=%r'%compression
            return wcs
        except Exception as err:
            pass
    else:
        # Finally, this one is really the last resort, since it only reads in the linear part of the
        # WCS.  It defaults to the equivalent of a pixel scale of 1.0 if even these are not present.
        if not suppress_warning:
            galsim_warn("All the fits WCS types failed to read %r. Using AffineTransform "
                        "instead, which will not really be correct."%(file_name))
        return AffineTransform._readHeader(header)

# Let this function work like a class in config.
FitsWCS._req_params = { "file_name" : str }
FitsWCS._opt_params = { "dir" : str, "hdu" : int, "compression" : str, 'text_file' : bool }

def FittedSIPWCS(x, y, ra, dec, wcs_type='TAN', order=3, center=None):
    """A WCS constructed from a list of reference celestial and image
    coordinates.

    Parameters:
        x:         Image x-coordinates of reference stars in pixels
        y:         Image y-coordinates of reference stars in pixels
        ra:        Right ascension of reference stars in radians
        dec:       Declination of reference stars in radians
        wcs_type:  The type of the tangent plane projection to use.  Should be
                   one of ['TAN', 'STG', 'ZEA', or 'ARC'].  [default: 'TAN']
        order:     The order of the Simple Imaging Polynomial (SIP) used to
                   describe the WCS distortion.  SIP coefficients kick in when
                   order >= 2.  If you supply order=1, then just fit a WCS
                   without any SIP coefficients.  [default: 3]
        center:    A `CelestialCoord` defining the location on the sphere where
                   the tangent plane is centered.  [default: None, which means
                   use the average position of the list of reference stars]
    """
    if order < 1:
        raise GalSimValueError("Illegal SIP order", order)

    nstar = len(x)
    # Make sure we have enough stars.
    # We need 1 star for crpix, 2 more for cd, and then
    # (order+2)*(order+1)/2 - 3 for ab.  The total is then (order+1)*(order+2)/2
    nrequire = (order+1)*(order+2)/2
    if nstar < nrequire:
        raise GalSimError(
            "Require at least {:0} stars for SIP order {:1}"
            .format(nrequire, order)
        )

    if center is None:
        # Use deprojected 3D mean of ra/dec unit sphere points as center
        wx = np.mean(np.cos(dec)*np.cos(ra))
        wy = np.mean(np.cos(dec)*np.sin(ra))
        wz = np.mean(np.sin(dec))
        center = CelestialCoord.from_xyz(wx, wy, wz)

    # Project radec onto uv so we can linearly fit the CRPIX and CD matrix
    # initial guesses
    u, v = center.project_rad(ra, dec)
    a = np.array(np.broadcast_arrays(1., -u, v)).T
    b = np.array([x, y]).T
    r, _, _, _ = np.linalg.lstsq(a, b, rcond=None)
    crpix_guess = r[0]
    cd_guess = np.linalg.inv(np.deg2rad(r[1:]))

    # SIP coefficient initial guesses are just 0.0
    ab_guess = []
    for i in range(order+1):
        for j in range(order+1):
            if (i+j > 1) and (i+j <= order):
                ab_guess.extend([0.0, 0.0])
    ab_guess = np.array(ab_guess)
    guess = np.hstack([crpix_guess, cd_guess.ravel(), ab_guess.ravel()])

    def _getWCS(wcs_type, center, crpix, cd, ab=None, abp=None, doiter=False):
        _data = [
            wcs_type if order == 1 else wcs_type+'-SIP',
            crpix,
            cd,
            center,
            None, # pv, unused
            ab,
            abp
        ]
        return GSFitsWCS(_data=_data, _doiter=doiter)

    def _decodeSIP(order, params, min=2):
        if order == 1:
            return None
        k = 0
        a = []
        b = []
        for i in range(order+1):
            for j in range(order+1):
                if (i+j < min) or (i+j > order):
                    a.append(0.0)
                    b.append(0.0)
                else:
                    a.append(params[k])
                    b.append(params[k+1])
                    k += 2
        a = np.array(a).reshape((order+1, order+1))
        b = np.array(b).reshape((order+1, order+1))
        a[1,0] += 1  # GSFitsWCS wants these with the identity included.
        b[0,1] += 1
        ab = np.array([a, b])
        return ab

    def _abLoss(params, order, wcs_type, center, x, y, u, v):
        crpix = params[:2]
        cd = params[2:6].reshape(2, 2)
        ab = _decodeSIP(order, params[6:])
        wcs = _getWCS(wcs_type, center, crpix, cd, ab)
        ra_p, dec_p = wcs.xyToradec(x, y, units='rad')
        u_p, v_p = center.project_rad(ra_p, dec_p)
        resid = np.hstack([u-u_p, v-v_p])
        resid = np.rad2deg(resid)*3600*1e6  # Work in microarcseconds
        return resid

    result = least_squares(
        _abLoss,
        guess,
        args=(order, wcs_type, center, x, y, u, v)
    )
    # rmse = np.sqrt(2*result.cost/len(u))
    # print(f"rmse: {rmse:.2f} microarcsec")
    crpix = result.x[0:2]
    cd = result.x[2:6].reshape(2, 2)
    ab = _decodeSIP(order, result.x[6:])

    if order == 1:
        return _getWCS(wcs_type, center, crpix, cd)

    # Now go back holding crpix, cd, and ab constant, and solve for inverse
    # coefficients abp
    # ABP SIP coefficient initial guesses are just 0.0
    abp_guess = []
    for i in range(order+1):
        for j in range(order+1):
            if (i+j > 0) and (i+j <= order):
                abp_guess.extend([0.0, 0.0])
    abp_guess = np.array(abp_guess)

    def _abpLoss(params, order, wcs_type, center, crpix, cd, ab, ra, dec, x, y):
        abp = _decodeSIP(order, params, min=1)
        wcs = _getWCS(wcs_type, center, crpix, cd, ab, abp)
        x_p, y_p = wcs.radecToxy(ra, dec, units='rad')
        resid = np.hstack([x-x_p, y-y_p])*1e6  # work in micropixels
        return resid
    result = least_squares(
        _abpLoss, abp_guess,
        args=(order, wcs_type, center, crpix, cd, ab, ra, dec, x, y)
    )
    # rmse = np.sqrt(2*result.cost/len(u))
    # print(f"rmse: {rmse:.2f} micropixels")
    abp = _decodeSIP(order, result.x, min=1)
    return _getWCS(wcs_type, center, crpix, cd, ab, abp, doiter=True)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import logging
from astropy.units import Quantity, Unit

from .util import LoggerWrapper
from .value import ParseValue, GetAllParams, GetIndex
from .input import RegisterInputConnectedType
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..bandpass import Bandpass
from ..utilities import basestring

# This module-level dict will store all the registered Bandpass types.
# See the RegisterBandpassType function at the end of this file.
# The keys are the (string) names of the Bandpass types, and the values will be builders that know
# how to build the Bandpass object.
valid_bandpass_types = {}

def BuildBandpass(config, key, base, logger=None):
    """Read the Bandpass parameters from config[key] and return a constructed Bandpass object.

    Parameters:
        config:     A dict with the configuration information.
        key:        The key name in config indicating which object to build.
        base:       The base dict of the configuration.
        logger:     Optionally, provide a logger for logging debug statements. [default: None]

    Returns:
        (bandpass, safe) where bandpass is a Bandpass instance, and safe is whether it is safe to
                         reuse.
    """
    logger = LoggerWrapper(logger)
    logger.debug('obj %d: Start BuildBandpass key = %s',base.get('obj_num',0),key)

    param = config[key]

    # Check for direct value, else get the bandpass type
    if isinstance(param, Bandpass):
        return param, True
    elif isinstance(param, basestring) and (param[0] == '$' or param[0] == '@'):
        return ParseValue(config, key, base, None)
    elif isinstance(param, dict):
        bandpass_type = param.get('type', 'FileBandpass')
    else:
        raise GalSimConfigError("%s must be either a Bandpass or a dict"%key)

    # For these two, just do the usual ParseValue function.
    if bandpass_type in ('Eval', 'Current'):
        return ParseValue(config, key, base, None)

    # Check if we can use the current cached object
    index, index_key = GetIndex(param, base)
    if 'current' in param:
        cbandpass, csafe, cvalue_type, cindex, cindex_key = param['current']
        if cindex == index:
            logger.debug('obj %d: The Bandpass object is already current', base.get('obj_num',0))
            logger.debug('obj %d: index_key = %s, index = %d',base.get('obj_num',0),
                         cindex_key, cindex)
            return cbandpass, csafe

    if bandpass_type not in valid_bandpass_types:
        raise GalSimConfigValueError("Invalid bandpass.type.", bandpass_type,
                                     list(valid_bandpass_types.keys()))
    logger.debug('obj %d: Building bandpass type %s', base.get('obj_num',0), bandpass_type)
    builder = valid_bandpass_types[bandpass_type]
    bandpass, safe = builder.buildBandpass(param, base, logger)
    logger.debug('obj %d: bandpass = %s', base.get('obj_num',0), bandpass)

    param['current'] = bandpass, safe, Bandpass, index, index_key

    return bandpass, safe


class BandpassBuilder:
    """A base class for building Bandpass objects.

    The base class defines the call signatures of the methods that any derived class should follow.
    """
    def buildBandpass(self, config, base, logger):
        """Build the Bandpass based on the specifications in the config dict.

        Note: Sub-classes must override this function with a real implementation.

        Parameters:
            config:     The configuration dict for the bandpass type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed Bandpass object.
        """
        raise NotImplementedError("The %s class has not overridden buildBandpass"%self.__class__)


class FileBandpassBuilder(BandpassBuilder):
    """A class for loading a Bandpass from a file

    FileBandpass expects the following parameters:

        file_name (str)                 The file to load (required)
        wave_type (str or Quantity)     The units (nm or Ang) of the wavelengths in the file (required)
        thin (float)                    A relative error to use for thinning the file (default: None)
        blue_limit (float or Quantity)  A cutoff wavelength on the blue side (default: None)
        red_limit (float or Quantity)   A cutoff wavelength on the red side (default: None)
        zeropoint (float or str)        A zeropoint to use (default: None)
    """
    def buildBandpass(self, config, base, logger):
        """Build the Bandpass based on the specifications in the config dict.

        Parameters:
            config:     The configuration dict for the bandpass type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed Bandpass object.
        """
        logger = LoggerWrapper(logger)

        req = {
            'file_name': str,
            'wave_type': (str, Unit),
        }
        opt = {
            'thin' : float,
            'blue_limit' : (float, Quantity),
            'red_limit' : (float, Quantity),
            'zeropoint': (float, str)
        }

        kwargs, safe = GetAllParams(config, base, req=req, opt=opt)

        file_name = kwargs.pop('file_name')
        thin = kwargs.pop('thin', None)
        zeropoint = kwargs.pop('zeropoint', None)

        logger.info("Reading Bandpass file: %s",file_name)
        bandpass = Bandpass(file_name, **kwargs)
        if zeropoint:
            bandpass = bandpass.withZeropoint(zeropoint)
        if thin:
            bandpass = bandpass.thin(thin)

        return bandpass, safe

def RegisterBandpassType(bandpass_type, builder, input_type=None):
    """Register a bandpass type for use by the config apparatus.

    Parameters:
        bandpass_type:  The name of the type in the config dict.
        builder:        A builder object to use for building the Bandpass object.  It should
                        be an instance of a subclass of BandpassBuilder.
        input_type:     If the Bandpass builder utilises an input object, give the key name of the
                        input type here.  (If it uses more than one, this may be a list.)
                        [default: None]
    """
    valid_bandpass_types[bandpass_type] = builder
    RegisterInputConnectedType(input_type, bandpass_type)

RegisterBandpassType('FileBandpass', FileBandpassBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import os
import sys
import logging
import shutil
from unittest import mock
import galsim
import galsim.download_cosmos  # Not imported automatically
from galsim_test_helpers import *

# This file tests the galsim_download_cosmos executable.
# We use a lock of mocking to do these tests.  We mock user input, urlopen, and more.
# The directory fake_cosmos has a tarball for the 23.5 sample and a directory for 25.2.
# So we use these separately at times when we need one or the other to exist.

@timer
def test_args():
    """Test the argument parsing
    """
    # -h prints description and exits
    print('This should print the description...')  # Can only check by inspection, not assert
    with assert_raises(SystemExit):
        args = galsim.download_cosmos.parse_args(['-h'])

    # Check defaults:
    args = galsim.download_cosmos.parse_args([])
    assert args.verbosity == 2
    assert args.force is False
    assert args.quiet is False
    assert args.unpack is False
    assert args.save is False
    assert args.dir is None
    assert args.sample == '25.2'
    assert args.nolink is False

    # Check setting each of those to be something else
    args = galsim.download_cosmos.parse_args(['-v', '3'])
    assert args.verbosity == 3
    args = galsim.download_cosmos.parse_args(['-v', '0'])
    assert args.verbosity == 0

    args = galsim.download_cosmos.parse_args(['-f'])
    assert args.force is True

    args = galsim.download_cosmos.parse_args(['-q'])
    assert args.quiet is True

    args = galsim.download_cosmos.parse_args(['-u'])
    assert args.unpack is True

    args = galsim.download_cosmos.parse_args(['--save'])
    assert args.save is True

    args = galsim.download_cosmos.parse_args(['-d','~/share'])
    assert args.dir == '~/share'

    args = galsim.download_cosmos.parse_args(['-s','23.5'])
    assert args.sample == '23.5'
    args = galsim.download_cosmos.parse_args(['-s','25.2'])
    assert args.sample == '25.2'

    args = galsim.download_cosmos.parse_args(['--nolink'])
    assert args.nolink is True

    # Some invalid parameters
    # To avoid ugly text output during pytest runs, redirect stderr to stdout for a moment.
    sys_stderr = sys.stderr
    sys.stderr = sys.stdout
    with assert_raises(SystemExit):
        galsim.download_cosmos.parse_args(['-s', '25.9'])
    with assert_raises(SystemExit):
        galsim.download_cosmos.parse_args(['-v', '-1'])
    with assert_raises(SystemExit):
        galsim.download_cosmos.parse_args(['-v', '4'])
    sys.stderr = sys_stderr

# global for the bleh and delay functions
count = 0

@timer
def test_query():
    """Test the query_yes_no function

    Need to mock the input function for this
    """
    from galsim.download_cosmos import query_yes_no

    def bleh():
        global count
        count += 1
        return 'y' if count % 5 == 0 else 'bleh'

    def delay_y():
        global count
        count += 1
        return 'y' if count % 5 == 0 else ''

    def delay_n():
        global count
        count += 1
        return 'n' if count % 5 == 0 else ''

    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        assert query_yes_no('', 'yes') == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', return_value='yes'):
        assert query_yes_no('', 'yes') == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        assert query_yes_no('', 'yes') == 'no'
    with mock.patch('galsim.download_cosmos.get_input', return_value='no'):
        assert query_yes_no('', 'yes') == 'no'
    with mock.patch('galsim.download_cosmos.get_input', return_value=''):
        assert query_yes_no('', 'yes') == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', bleh):
        assert query_yes_no('', 'yes') == 'yes'

    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        assert query_yes_no('', 'no') == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', return_value='yes'):
        assert query_yes_no('', 'no') == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        assert query_yes_no('', 'no') == 'no'
    with mock.patch('galsim.download_cosmos.get_input', return_value='no'):
        assert query_yes_no('', 'no') == 'no'
    with mock.patch('galsim.download_cosmos.get_input', return_value=''):
        assert query_yes_no('', 'no') == 'no'
    with mock.patch('galsim.download_cosmos.get_input', bleh):
        assert query_yes_no('', 'yes') == 'yes'

    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        assert query_yes_no('', None) == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', return_value='yes'):
        assert query_yes_no('', None) == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        assert query_yes_no('', None) == 'no'
    with mock.patch('galsim.download_cosmos.get_input', return_value='no'):
        assert query_yes_no('', None) == 'no'
    with mock.patch('galsim.download_cosmos.get_input', delay_n):
        assert query_yes_no('', None) == 'no'
    with mock.patch('galsim.download_cosmos.get_input', delay_y):
        assert query_yes_no('', None) == 'yes'
    with mock.patch('galsim.download_cosmos.get_input', bleh):
        assert query_yes_no('', None) == 'yes'

    with assert_raises(ValueError):
        query_yes_no('', 'invalid')

# Need to call these before each time make_logger is repeated.  Else duplicate handles.
def remove_handler():
    logger = logging.getLogger('galsim')
    for handler in logger.handlers:
        logger.removeHandler(handler)

@timer
def test_names():
    """Test the get_names function
    """
    from galsim.download_cosmos import get_names

    args = galsim.download_cosmos.parse_args([])
    remove_handler()
    logger = galsim.download_cosmos.make_logger(args)
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)
    assert url == 'https://zenodo.org/record/3242143/files/COSMOS_25.2_training_sample.tar.gz'
    assert target_dir == galsim.meta_data.share_dir
    assert do_link is False
    assert target == os.path.join(target_dir, 'COSMOS_25.2_training_sample.tar.gz')
    assert unpack_dir == os.path.join(target_dir, 'COSMOS_25.2_training_sample')
    assert link_dir == os.path.join(galsim.meta_data.share_dir, 'COSMOS_25.2_training_sample')

    args = galsim.download_cosmos.parse_args(['-d','~/share','-s','23.5'])
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)
    assert url == 'https://zenodo.org/record/3242143/files/COSMOS_23.5_training_sample.tar.gz'
    assert target_dir == os.path.expanduser('~/share')
    assert do_link is True
    assert target == os.path.join(target_dir, 'COSMOS_23.5_training_sample.tar.gz')
    assert unpack_dir == os.path.join(target_dir, 'COSMOS_23.5_training_sample')
    assert link_dir == os.path.join(galsim.meta_data.share_dir, 'COSMOS_23.5_training_sample')

    args = galsim.download_cosmos.parse_args(['-d','share','--nolink'])
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)
    assert url == 'https://zenodo.org/record/3242143/files/COSMOS_25.2_training_sample.tar.gz'
    assert target_dir == 'share'
    assert do_link is False
    assert target == os.path.join(target_dir, 'COSMOS_25.2_training_sample.tar.gz')
    assert unpack_dir == os.path.join(target_dir, 'COSMOS_25.2_training_sample')
    assert link_dir == os.path.join(galsim.meta_data.share_dir, 'COSMOS_25.2_training_sample')


class fake_urlopen:
    err = None

    # We don't want to actually check things on the internet.  So this class fakes
    # up the real urlopen function.
    def __init__(self, url):
        self.n = 10
        pass
    def info(self):
        return {
            "Server": "nginx/1.16.1",
            "Content-Type": "application/octet-stream",
            "Content-Length": "728",
            "Connection": "close",
            "Content-MD5": "e05cfe60c037c645d61ac70545cc2a99",
            "Content-Security-Policy": "default-src 'none';",
            "X-Content-Type-Options": "nosniff",
            "X-Download-Options": "noopen",
            "X-Permitted-Cross-Domain-Policies": "none",
            "X-Frame-Options": "sameorigin",
            "X-XSS-Protection": "1; mode=block",
            "Content-Disposition": "attachment; filename=COSMOS_25.2_training_sample.tar.gz",
            "ETag": "\"md5:e05cfe60c037c645d61ac70545cc2a99\"",
            "Last-Modified": "Sun, 31 May 2020 02:19:18 GMT",
            "Date": "Thu, 11 Jun 2020 16:06:07 GMT",
            "Accept-Ranges": "none",
            "X-RateLimit-Limit": "60",
            "X-RateLimit-Remaining": "59",
            "X-RateLimit-Reset": "1591891628",
            "Retry-After": "60",
            "Strict-Transport-Security": "max-age=0",
            "Referrer-Policy": "strict-origin-when-cross-origin",
            "Set-Cookie": "session=3765c5a1d5211e53_5ee2566f",
            "X-Session-ID": "3765c5a1d5211e53_5ee2566f",
            "X-Request-ID": "2b720f14bdd71a29031a5cb415b391f8"
        }
    def read(self, block_sz):
        if self.err:
            raise OSError(self.err)
        if self.n:
            self.n -= 1
            return b'x' * 80  # ignore block_sz
        else:
            return b''

@timer
def test_check():
    """Test the get_meta and check_existing functions

    The latter of these is really the most interesting, and has the most bits
    of anything in the script that are worth checking with unit tests.
    """
    from galsim.download_cosmos import get_names, get_meta, check_existing

    args = galsim.download_cosmos.parse_args(['-d','fake_cosmos','-q','-v','3'])
    remove_handler()
    logger = galsim.download_cosmos.make_logger(args)
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)

    # Check get_meta
    with mock.patch('galsim.download_cosmos.urlopen', fake_urlopen):
        meta = get_meta(url, args, logger)
        assert meta['Content-Length'] == "728"
        assert meta['Content-MD5'] == "e05cfe60c037c645d61ac70545cc2a99"

    # File already exists and is current.
    do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is False

    # Some changes imply it's obsolete
    meta['Server'] =  "nginx/1.23.1"
    meta['X-Content-Type-Options'] = "sniff"
    meta['Last-Modified'] = "Tue, 12 Mar 2019 08:12:12 GMT"
    meta['Date'] = "Sun, 14 Jun 2020 20:00:00 GMT"
    meta['X-RateLimit-Remaining'] = "31"
    meta['Retry-After'] = "120"
    meta['Set-Cookie'] =  "session=2b720f14bdd71a29031a5cb415b391f8"
    do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is False

    # Force download anyway
    args.quiet = False
    args.force = True
    do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is True

    # Ask whether to re-download
    args.force = False
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is True
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is False

    # Meta data is obsolete
    meta1 = meta.copy()
    meta1['Content-Length'] = "9999"
    meta1['Content-MD5'] = "f05cfe60c037c645d61ac70545cc2a99"
    args.quiet = True
    do_download = check_existing(target, unpack_dir, meta1, args, logger)
    assert do_download is True

    # If they change the name of the checksum key, we consider it obsolete.
    meta2 = meta1.copy()
    meta2['Content-New-MD5'] = "e05cfe60c037c645d61ac70545cc2a99"
    del meta2['Content-MD5']
    del meta2['ETag']
    do_download = check_existing(target, unpack_dir, meta2, args, logger)
    assert do_download is True

    args.quiet = False
    args.force = True
    do_download = check_existing(target, unpack_dir, meta1, args, logger)
    assert do_download is True

    args.force = False
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        do_download = check_existing(target, unpack_dir, meta1, args, logger)
    assert do_download is True
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        do_download = check_existing(target, unpack_dir, meta1, args, logger)
    assert do_download is False

    # Meta data is missing
    args.quiet = True
    do_download = check_existing(target, 'output', meta, args, logger)
    assert do_download is True

    # Tarball is present, but wrong size
    args = galsim.download_cosmos.parse_args(['-d','fake_cosmos','-s','23.5','-q'])
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)
    do_download = check_existing(target, 'output', meta1, args, logger)
    assert do_download is True

    args.quiet = False
    args.force = True
    do_download = check_existing(target, 'output', meta1, args, logger)
    assert do_download is True

    args.force = False
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        do_download = check_existing(target, unpack_dir, meta1, args, logger)
    assert do_download is True
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        do_download = check_existing(target, unpack_dir, meta1, args, logger)
    assert do_download is False

    # Tarball is present, and correct size
    args.quiet = True
    do_download = check_existing(target, 'output', meta, args, logger)
    assert do_download is False

    args.quiet = False
    args.force = True
    do_download = check_existing(target, 'output', meta, args, logger)
    assert do_download is True

    args.force = False
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is True
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is False

    # Tarball and unpack_dir both missing
    args = galsim.download_cosmos.parse_args(['-d','input'])
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)
    do_download = check_existing(target, unpack_dir, meta, args, logger)
    assert do_download is True


@timer
def test_download():
    """Test the download function

    This one is a little silly.  It's almost completely mocked.  But we can at least check
    that there are no bugs that would raise an exception of some sort.
    """
    from galsim.download_cosmos import get_names, get_meta, download

    args = galsim.download_cosmos.parse_args(['-d','output','-q'])
    remove_handler()
    logger = galsim.download_cosmos.make_logger(args)
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)

    with mock.patch('galsim.download_cosmos.urlopen', fake_urlopen):
        meta = get_meta(url, args, logger)

        print('Start download with verbosity = 2')
        download(True, url, target, meta, args, logger)

        print('Start download with verbosity = 1')
        args.verbosity = 1
        download(True, url, target, meta, args, logger)

        print('Start download with verbosity = 3')
        args.verbosity = 3
        download(True, url, target, meta, args, logger)

        print('Start download with verbosity = 0')
        args.verbosity = 0
        download(True, url, target, meta, args, logger)

        print("Don't download")
        download(False, url, target, meta, args, logger)

        fake_urlopen.err = 'Permission denied'
        with CaptureLog() as cl:
            assert_raises(OSError, download, True, url, target, meta, args, cl.logger)
        assert "Rerun using sudo" in cl.output

        fake_urlopen.err = 'Disk quota exceeded'
        with CaptureLog() as cl:
            assert_raises(OSError, download, True, url, target, meta, args, cl.logger)
        assert "You might need to download this in an alternate location" in cl.output

        fake_urlopen.err = 'gack'
        assert_raises(OSError, download, True, url, target, meta, args, logger)
    fake_urlopen.err = None

@timer
def test_unpack():
    """Test the check_unpack and unpack functions
    """
    from galsim.download_cosmos import get_names, check_unpack, unpack

    # If we downloaded the file, then we usually want to unpack
    args = galsim.download_cosmos.parse_args(['-d','fake_cosmos','-s','23.5','-q'])
    remove_handler()
    logger = galsim.download_cosmos.make_logger(args)
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)
    meta = fake_urlopen(url).info()

    print('unpack_dir = ',unpack_dir)
    if os.path.exists(unpack_dir):
        shutil.rmtree(unpack_dir)

    # Regular case, downloaded file and not unpacked yet
    do_unpack = check_unpack(True, unpack_dir, target, args)
    assert do_unpack is True

    # If we didn't download, but tarball exists, still unpack
    do_unpack = check_unpack(False, unpack_dir, target, args)
    assert do_unpack is True

    # Now unpack it
    print('unpack with verbose = 2:')
    unpack(True, target, target_dir, unpack_dir, meta, args, logger)

    shutil.rmtree(unpack_dir)
    print('unpack with verbose = 3:')
    args.verbosity = 3
    unpack(True, target, target_dir, unpack_dir, meta, args, logger)

    shutil.rmtree(unpack_dir)
    print('unpack with verbose = 1:')
    args.verbosity = 1
    unpack(True, target, target_dir, unpack_dir, meta, args, logger)

    print("Don't unpack")
    unpack(False, target, target_dir, unpack_dir, meta, args, logger)

    # If it is already unpacked, probably don't unpack it
    do_unpack = check_unpack(False, unpack_dir, target, args)
    assert do_unpack is False

    # Unless we expressly say to on the command line
    args.unpack = True
    do_unpack = check_unpack(False, unpack_dir, target, args)
    assert do_unpack is True

    # Or if we downloaded a new tarball, we will unpack it.
    args.unpack = False
    do_unpack = check_unpack(True, unpack_dir, target, args)
    assert do_unpack is True

    # Or if not quiet, it will ask whether to re-unpack the tarball
    args.quiet = False
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        do_unpack = check_unpack(False, unpack_dir, target, args)
    assert do_unpack is True
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        do_unpack = check_unpack(False, unpack_dir, target, args)
    assert do_unpack is False

    # Finally, if the tarball doesn't exist, then we can't unpack it.
    target2 = target.replace('23.5','25.2')
    do_unpack = check_unpack(True, unpack_dir, target2, args)
    assert do_unpack is False

@timer
def test_remove():
    """Test the check_remove and remove_tarball function
    """
    from galsim.download_cosmos import get_names, check_remove, remove_tarball

    args = galsim.download_cosmos.parse_args(['-d','fake_cosmos','-s','23.5','-q'])
    remove_handler()
    logger = galsim.download_cosmos.make_logger(args)
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)

    # Normally, we remove the tarball if we unpacked it.
    do_remove = check_remove(True, target, args)
    assert do_remove is True

    # Or if we say to save it, then save it
    args.save = True
    do_remove = check_remove(True, target, args)
    assert do_remove is False

    # If we didn't unpack it, probably don't delete it
    args.save = False
    do_remove = check_remove(False, target, args)
    assert do_remove is False

    # But ask if not quiet to be sure
    args.quiet = False
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        do_remove = check_remove(False, target, args)
    assert do_remove is True
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        do_remove = check_remove(False, target, args)
    assert do_remove is False

    target1 = target + '.tar'
    with open(target1,'w') as f:
        f.write('blah')
    assert os.path.isfile(target1)
    remove_tarball(False, target1, logger)
    assert os.path.isfile(target1)
    remove_tarball(True, target1, logger)
    assert not os.path.isfile(target1)


@timer
def test_link():
    """Test the link_cosmos function
    """
    from galsim.download_cosmos import get_names, make_link

    args = galsim.download_cosmos.parse_args(['-d','fake_cosmos','-q'])
    remove_handler()
    logger = galsim.download_cosmos.make_logger(args)
    url, target, target_dir, link_dir, unpack_dir, do_link = get_names(args, logger)
    link_dir = os.path.join('output', 'COSMOS_25.2_training_sample')

    # If link doesn't exist yet, make it.
    if os.path.lexists(link_dir):
        os.unlink(link_dir)
    assert not os.path.lexists(link_dir)
    make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.islink(link_dir)

    # If link already exists, remove and remake
    make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.islink(link_dir)

    # If do_link is False, don't make it
    os.unlink(link_dir)
    make_link(False, unpack_dir, link_dir, args, logger)
    assert not os.path.exists(link_dir)

    # If link exists, but is a directory, don't remove it
    os.mkdir(link_dir)
    assert os.path.isdir(link_dir)
    make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.isdir(link_dir)
    assert not os.path.islink(link_dir)

    # Unless force
    args.force = True
    make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.islink(link_dir)

    # Or ask
    args.force = False
    args.quiet = False
    os.unlink(link_dir)
    os.mkdir(link_dir)
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.isdir(link_dir)
    assert not os.path.islink(link_dir)
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.islink(link_dir)

    # If it's a broken link, remove and relink
    os.unlink(link_dir)
    os.symlink('invalid', link_dir)
    assert os.path.lexists(link_dir)
    assert os.path.islink(link_dir)
    assert not os.path.exists(link_dir)
    make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.islink(link_dir)

    # If it's a file, probably remove and relink
    os.unlink(link_dir)
    with open(link_dir,'w') as f: f.write('blah')
    args.quiet = True
    assert os.path.exists(link_dir)
    assert not os.path.islink(link_dir)
    make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.islink(link_dir)

    os.unlink(link_dir)
    with open(link_dir,'w') as f: f.write('blah')
    args.quiet = False
    args.force = True
    assert os.path.exists(link_dir)
    assert not os.path.islink(link_dir)
    make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.islink(link_dir)

    # But ask if appropriate
    os.unlink(link_dir)
    with open(link_dir,'w') as f: f.write('blah')
    args.force = False
    assert os.path.exists(link_dir)
    assert not os.path.islink(link_dir)
    with mock.patch('galsim.download_cosmos.get_input', return_value='n'):
        make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert not os.path.islink(link_dir)
    with mock.patch('galsim.download_cosmos.get_input', return_value='y'):
        make_link(True, unpack_dir, link_dir, args, logger)
    assert os.path.exists(link_dir)
    assert os.path.islink(link_dir)

@timer
def test_full():
    """Test the full script
    """
    link_dir1 = os.path.join('output', 'COSMOS_23.5_training_sample')
    link_dir2 = os.path.join('output', 'COSMOS_25.2_training_sample')
    if os.path.lexists(link_dir1):
        os.unlink(link_dir1)
    if os.path.lexists(link_dir2):
        os.unlink(link_dir2)

    with mock.patch('galsim.download_cosmos.share_dir', 'output'), \
         mock.patch('galsim.download_cosmos.urlopen', fake_urlopen):

        remove_handler()
        assert not os.path.islink(link_dir1)
        galsim.download_cosmos.main(['-d','fake_cosmos','-q','-s','23.5','--save'])
        assert os.path.islink(link_dir1)

        remove_handler()
        assert not os.path.islink(link_dir2)
        with mock.patch('sys.argv', ['galsim_download_cosmos', '-d', 'fake_cosmos', '-q']):
            galsim.download_cosmos.run_main()
        assert os.path.islink(link_dir2)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'GSParams' ]

import copy

from . import _galsim

class GSParams:
    """GSParams stores a set of numbers that govern how a `GSObject` makes various speed/accuracy
    tradeoff decisions.

    All `GSObject` classes can take an optional parameter named ``gsparams``, which would be an
    instance of this class.  e.g.::

        >>> gsp = galsim.GSParams(folding_threshold=1.e-3)
        >>> gal = galsim.Sersic(n=3.4, half_light_radius=3.2, flux=200, gsparams=gsp)

    One can also update the parameters for an existing object using the method
    `GSObject.withGSParams`.  e.g.::

        >>> gal = gal.withGSParams(kvalue_accuracy=1.e-8)

    All parameters have reasonable default values.  You only need to specify the ones you want
    to change.

    Parameters:
        minimum_fft_size:   The minimum size of any FFT that may need to be performed.
                            [default: 128]
        maximum_fft_size:   The maximum allowed size of an image for performing an FFT without
                            warning.  This is more about memory use than accuracy.  We have this
                            maximum value to inform a user who accidentally performs an extremely
                            large FFT why they just crashed the program. GalSim used to
                            raise an exception indicating that the image is too large,
                            which is often a sign of an error in the user's code. However, we
                            now just emit a warning about the large FFT, so if the code crashes
                            you have some indication of why. If you have the memory to handle it,
                            you can raise this limit to allow the calculation to happen without
                            seeing the warning. [default: 8192]

                            .. note::

                                To revert to the old behavior of raising an exception, you
                                may set ``galsim.errors.raise_fft_size_error = True``.

        folding_threshold:  This sets a maximum amount of real space folding that is allowed,
                            an effect caused by the periodic nature of FFTs.  FFTs implicitly
                            use periodic boundary conditions, and a profile specified on a
                            finite grid in Fourier space corresponds to a real space image
                            that will have some overlap with the neighboring copies of the real
                            space profile.  As the step size in k increases, the spacing
                            between neighboring aliases in real space decreases, increasing the
                            amount of folded, overlapping flux.  ``folding_threshold`` is used
                            to set an appropriate step size in k to allow at most this
                            fraction of the flux to be folded.
                            This parameter is also relevant when you let GalSim decide how
                            large an image to use for your object.  The image is made to be
                            large enough that at most a fraction ``folding_threshold`` of the
                            total flux is allowed to fall off the edge of the image.
                            [default: 5.e-3]
        stepk_minimum_hlr:  In addition to the above constraint for aliasing, also set stepk
                            such that pi/stepk is at least ``stepk_minimum_hlr`` times the
                            profile's half-light radius (for profiles that have a well-defined
                            half-light radius). [default: 5]
        maxk_threshold:     This sets the maximum amplitude of the high frequency modes in
                            Fourier space that are excluded by truncating the FFT at some
                            maximum k value. Lowering this parameter can help minimize the
                            effect of "ringing" if you see that in your images.
                            [default: 1.e-3]
        kvalue_accuracy:    This sets the accuracy of values in Fourier space. Whenever there
                            is some kind of approximation to be made in the calculation of a
                            Fourier space value, the error in the approximation is constrained
                            to be no more than this value times the total flux.
                            [default: 1.e-5]
        xvalue_accuracy:    This sets the accuracy of values in real space. Whenever there is
                            some kind of approximation to be made in the calculation of a
                            real space value, the error in the approximation is constrained
                            to be no more than this value times the total flux.
                            [default: 1.e-5]
        table_spacing:      Several profiles use lookup tables for either the Hankel transform
                            (`Sersic`, `Moffat`) or the real space radial function (`Kolmogorov`).
                            We try to estimate a good spacing between values in the lookup
                            tables based on either ``xvalue_accuracy`` or ``kvalue_accuracy`` as
                            appropriate. However, you may change the spacing with this
                            parameter. Using ``table_spacing < 1`` will use a spacing value that
                            is that much smaller than the default, which should produce more
                            accurate interpolations. [default: 1]
        realspace_relerr:   This sets the relative error tolerance for real-space integration.
                            [default: 1.e-4]
        realspace_abserr:   This sets the absolute error tolerance for real-space integration.
                            [default: 1.e-6]
                            The estimated integration error for the flux value in each pixel
                            when using the real-space rendering method (either explicitly with
                            ``method='real_space'`` or if it is triggered automatically with
                            ``method='auto'``) is constrained to be no larger than either
                            ``realspace_relerr`` times the pixel flux or ``realspace_abserr``
                            times the object's total flux.
        integration_relerr: The relative error tolerance for integrations other than real-space
                            rendering. [default: 1.e-6]
        integration_abserr: The absolute error tolerance for integrations other than real-space
                            rendering. [default: 1.e-8]
        shoot_accuracy:     This sets the relative accuracy on the total flux when photon
                            shooting.  The photon shooting algorithm at times needs to make
                            approximations, such as how high in radius it needs to sample the
                            radial profile. When such approximations need to be made, it makes
                            sure that the resulting fractional error in the flux will be at
                            most this much. [default: 1.e-5]

    After construction, all of the above parameters are available as read-only attributes.
    """
    def __init__(self, minimum_fft_size=128, maximum_fft_size=8192,
                 folding_threshold=5.e-3, stepk_minimum_hlr=5, maxk_threshold=1.e-3,
                 kvalue_accuracy=1.e-5, xvalue_accuracy=1.e-5, table_spacing=1,
                 realspace_relerr=1.e-4, realspace_abserr=1.e-6,
                 integration_relerr=1.e-6, integration_abserr=1.e-8,
                 shoot_accuracy=1.e-5, allowed_flux_variation=0.81,
                 range_division_for_extrema=32, small_fraction_of_flux=1.e-4):
        self._minimum_fft_size = int(minimum_fft_size)
        self._maximum_fft_size = int(maximum_fft_size)
        self._folding_threshold = float(folding_threshold)
        self._stepk_minimum_hlr = float(stepk_minimum_hlr)
        self._maxk_threshold = float(maxk_threshold)
        self._kvalue_accuracy = float(kvalue_accuracy)
        self._xvalue_accuracy = float(xvalue_accuracy)
        self._table_spacing = int(table_spacing)
        self._realspace_relerr = float(realspace_relerr)
        self._realspace_abserr = float(realspace_abserr)
        self._integration_relerr = float(integration_relerr)
        self._integration_abserr = float(integration_abserr)
        self._shoot_accuracy = float(shoot_accuracy)

        if allowed_flux_variation != 0.81:
            from .deprecated import depr
            depr('allowed_flux_variation', 2.1, "", "This parameter is no longer used.")
        if range_division_for_extrema != 32:
            from .deprecated import depr
            depr('range_division_for_extrema', 2.1, "", "This parameter is no longer used.")
        if small_fraction_of_flux != 1.e-4:
            from .deprecated import depr
            depr('small_fraction_of_flux', 2.1, "", "This parameter is no longer used.")

        self._gsp = _galsim.GSParams(*self._getinitargs())

    # Make all the attributes read-only
    @property
    def minimum_fft_size(self): return self._minimum_fft_size
    @property
    def maximum_fft_size(self): return self._maximum_fft_size
    @property
    def folding_threshold(self): return self._folding_threshold
    @property
    def stepk_minimum_hlr(self): return self._stepk_minimum_hlr
    @property
    def maxk_threshold(self): return self._maxk_threshold
    @property
    def kvalue_accuracy(self): return self._kvalue_accuracy
    @property
    def xvalue_accuracy(self): return self._xvalue_accuracy
    @property
    def table_spacing(self): return self._table_spacing
    @property
    def realspace_relerr(self): return self._realspace_relerr
    @property
    def realspace_abserr(self): return self._realspace_abserr
    @property
    def integration_relerr(self): return self._integration_relerr
    @property
    def integration_abserr(self): return self._integration_abserr
    @property
    def shoot_accuracy(self): return self._shoot_accuracy

    @staticmethod
    def check(gsparams, default=None, **kwargs):
        """Checks that gsparams is either a valid GSParams instance or None.

        In the former case, it returns gsparams, in the latter it returns default
        (GSParams.default if no other default specified).
        """
        if gsparams is None:
            gsparams = default if default is not None else GSParams.default
        elif not isinstance(gsparams, GSParams):
            raise TypeError("Invalid GSParams: %s"%gsparams)
        return gsparams.withParams(**kwargs)

    def withParams(self, **kwargs):
        """Return a `GSParams` that is identical to the current one except for any keyword
        arguments given here, which supersede the current value.
        """
        if len(kwargs) == 0:
            return self
        else:
            ret = copy.copy(self)
            for k in kwargs:
                if not hasattr(ret, '_' + k):
                    raise TypeError('parameter %s is invalid'%k)
                setattr(ret, '_' + k, kwargs[k])
            ret._gsp = _galsim.GSParams(*ret._getinitargs())
            return ret

    @staticmethod
    def combine(gsp_list):
        """Combine a list of `GSParams` instances using the most restrictive parameter from each.

        Uses the minimum value for most parameters. For the following parameters, it uses the
        maximum numerical value: minimum_fft_size, maximum_fft_size, stepk_minimum_hlr.
        """
        if len(gsp_list) == 1:
            return gsp_list[0]
        elif all(g == gsp_list[0] for g in gsp_list[1:]):
            return gsp_list[0]
        else:
            return GSParams(
                max([g.minimum_fft_size for g in gsp_list if g is not None]),
                max([g.maximum_fft_size for g in gsp_list if g is not None]),
                min([g.folding_threshold for g in gsp_list if g is not None]),
                max([g.stepk_minimum_hlr for g in gsp_list if g is not None]),
                min([g.maxk_threshold for g in gsp_list if g is not None]),
                min([g.kvalue_accuracy for g in gsp_list if g is not None]),
                min([g.xvalue_accuracy for g in gsp_list if g is not None]),
                min([g.table_spacing for g in gsp_list if g is not None]),
                min([g.realspace_relerr for g in gsp_list if g is not None]),
                min([g.realspace_abserr for g in gsp_list if g is not None]),
                min([g.integration_relerr for g in gsp_list if g is not None]),
                min([g.integration_abserr for g in gsp_list if g is not None]),
                min([g.shoot_accuracy for g in gsp_list if g is not None]))

    # Define once the order of args in __init__, since we use it a few times.
    def _getinitargs(self):
        return (int(self.minimum_fft_size), int(self.maximum_fft_size),
                self.folding_threshold, self.stepk_minimum_hlr, self.maxk_threshold,
                self.kvalue_accuracy, self.xvalue_accuracy, self.table_spacing,
                self.realspace_relerr, self.realspace_abserr,
                self.integration_relerr, self.integration_abserr,
                self.shoot_accuracy)

    def __getstate__(self): return self._getinitargs()
    def __setstate__(self, state): self.__init__(*state)

    def __repr__(self):
        return 'galsim.GSParams(%d,%d,%r,%r,%r,%r,%r,%d,%r,%r,%r,%r,%r)'% \
                self._getinitargs()

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, GSParams) and self._getinitargs() == other._getinitargs()))
    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(repr(self))

# We use the default a lot, so make it a class attribute.
GSParams.default = GSParams()
